// ALISP engine v0.4.3b
// ALISP is dedicated to Enota, the girl i'll never forget
// coded by Ketmar // Vampire Avalon (psyc://ketmar.no-ip.org/~Ketmar)
// Understanding is not required. Only obedience.
//
// This program is free software. It comes without any warranty, to
// the extent permitted by applicable law. You can redistribute it
// and/or modify it under the terms of the Do What The Fuck You Want
// To Public License, Version 2, as published by Sam Hocevar. See
// http://sam.zoy.org/wtfpl/COPYING for more details.
//
function ALPQuote (n: Integer): Integer;
begin
  result := n;
  // list of one item becomes this item itself %-)
  if (result <> al_Nil) and (al_CDR[result].vi = al_Nil) then
    result := al_CAR[result];
end;

function ALPXQuote (n: Integer): Integer;
begin
  result := n;
end;

function ALPThrow (n: Integer): Integer;
begin
  //Write('THROW ['); ALPrint(eng cdr[n]); WriteLn(']');
  if al_CDR[n].vi = al_Nil then result := al_CAR[n] else result := n;
  al_BreakFlag := ALBREAK_THROW;
  //Write('THROW <'); ALPrint(eng result); WriteLn('>');
end;

function ALPReturn (n: Integer): Integer;
begin
  result := al_Nil;
  while n <> al_Nil do
  begin
    result := al_CAR[n]; n := al_CDR[n].vi;
  end;
  al_BreakFlag := ALBREAK_RETURN;
end;

function ALPBreak (n: Integer): Integer;
begin
  result := al_Nil;
  while n <> al_Nil do
  begin
    result := al_CAR[n]; n := al_CDR[n].vi;
  end;
  al_BreakFlag := ALBREAK_BREAK;
end;

function ALPContinue (n: Integer): Integer;
begin
  result := al_Nil;
  while n <> al_Nil do
  begin
    result := al_CAR[n]; n := al_CDR[n].vi;
  end;
  al_BreakFlag := ALBREAK_CONT;
end;

function ALPEval (n: Integer): Integer;
var
  ctxc: Integer;
begin
  if al_CDR[n].vi = al_Nil then
  begin
    // one arg
    result := ALEval(al_CAR[n]);
    exit;
  end;

  // eval in new context...
  ctxc := al_CAR[al_CDR[n].vi]; // 2nd arg (context)
  // context?
  if not ALIsContext(ctxc) then
  begin
    result := ALThrowError('EVAL: context expected', ctxc);
    exit;
  end;
  result := al_CAR[n]; // 1st arg
  if not ALEnterContext(ctxc) then
  begin
    result := ALThrowError('EVAL: context error', ctxc);
    exit;
  end;
  result := ALEval(result);
  if al_BreakFlag = ALBREAK_ERROR then exit;
  ALLeaveContexts(1);
end;

function ALPFluid (n: Integer): Integer;
begin
  if al_Frames <> al_Nil then
  begin
    // local fluids should be processed by optimizer
    result := al_snil;
  end
  else result := DoFluidList(n, al_FluidList, true);
end;

function ALPUnfluid (n: Integer): Integer;
begin
  if al_Frames <> al_Nil then
  begin
    // local fluids should be processed by optimizer
    result := al_snil;
  end
  else result := DoFluidList(n, al_FluidList, false);
end;

function ALPLambda (n: Integer): Integer;
begin
  result := ALCons(al_slambda, n);
  if al_Optimize then
  begin
    result := LispCompile(result);
  end;
end;

function ALPMLambda (n: Integer): Integer;
begin
  result := ALCons(al_smlambda, n);
  if al_Optimize then
  begin
    result := LispCompile(result);
  end;
end;


procedure ALDefineBuiltInPrimitives ();
begin
  ALAddPrimitive('quote', ALPQuote, true, 1, -1);
  ALAddPrimitive('xquote', ALPXQuote, true, 1, -1);
  ALAddPrimitive('throw', ALPThrow, false, 1, -1);
  ALAddPrimitive('return', ALPReturn, false, 0, -1);
  ALAddPrimitive('break', ALPBreak, false, 0, -1);
  ALAddPrimitive('continue', ALPContinue, false, 0, -1);
  ALAddPrimitive('eval', ALPEval, false, 1, 2);
  ALAddPrimitive('fluid', ALPFluid, true, 1, -1);
  ALAddPrimitive('unfluid', ALPUnfluid, true, 1, -1);
  ALAddPrimitive('lambda', ALPLambda, true, 1, -1);
  ALAddPrimitive('mlambda', ALPMLambda, true, 1, -1);

  al_slambda := al_CDR[ALFindGlobalSymbol('lambda')].vi;
  al_smlambda := al_CDR[ALFindGlobalSymbol('mlambda')].vi;
  al_squote := al_CDR[ALFindGlobalSymbol('quote')].vi;
  al_sxquote := al_CDR[ALFindGlobalSymbol('xquote')].vi;
  al_sfluid := al_CDR[ALFindGlobalSymbol('fluid')].vi;
  al_sunfluid := al_CDR[ALFindGlobalSymbol('unfluid')].vi;
end;
