// ALISP engine v0.4.3b
// ALISP is dedicated to Enota, the girl i'll never forget
// coded by Ketmar // Vampire Avalon (psyc://ketmar.no-ip.org/~Ketmar)
// Understanding is not required. Only obedience.
//
// This program is free software. It comes without any warranty, to
// the extent permitted by applicable law. You can redistribute it
// and/or modify it under the terms of the Do What The Fuck You Want
// To Public License, Version 2, as published by Sam Hocevar. See
// http://sam.zoy.org/wtfpl/COPYING for more details.
//

//  [.] add conversion of byte-vector (and part of it) to string and vice versa
//  [.] add vector operations: to-list,

function ALPVecDelete (n: Integer): Integer;
// (vec-delete vector from [len])
// from<0: from the end of the vector
// no len: till the end of the vector
// returns vector
var
  vec, from, len, vlen: Integer;
  noLen: Boolean;
  tl: TList;
  tv: TLispByteVector;
begin
  if not ALNextAnyVectorArg('VEC-DELETE', n, vec, result) then exit;
  if not ALNextIntArg('VEC-DELETE', n, from, result) then exit;
  if ALIsNil(n) then noLen := true
  else
  begin
    noLen := false;
    if not ALNextIntArg('VEC-DELETE', n, len, result) then exit;
  end;
  result := vec;

  vlen := ALGetVectorLength(vec);
  if noLen then len := vlen;

  if from < 0 then from := vlen+from;
  if (from < 0) or (from >= vlen) then exit;
  if from+len > vlen then len := vlen-from;
  if len < 1 then exit;

  //WriteLn('from=', from, '; len=', len);
  if ALGetAtomType(vec) = DT_VECTOR then
  begin
    tl := TList(al_CDR[vec].vi);
    while len > 0 do begin tl.Delete(from); Dec(len); end;
  end
  else
  begin
    tv := TLispByteVector(al_CDR[vec].vi);
    while from+len < vlen do
    begin
      {$R+}
      tv.vector[from] := tv.vector[from+len]; Inc(from);
      {$R-}
    end;
    SetLength(tv.vector, vlen-len);
  end;
end;

function ALPVecExtract (n: Integer): Integer;
// (vec-extract vector from [len])
// from<0: from the end of the vector
// no len: till the end of the vector
// returns extracted vector (can be empty)
var
  vec, c, from, len, vlen: Integer;
  from0, len0: Integer;
  noLen: Boolean;
  tl, tl1: TList;
  tv, tv1: TLispByteVector;
begin
  if not ALNextAnyVectorArg('VEC-EXTRACT', n, vec, result) then exit;
  if not ALNextIntArg('VEC-EXTRACT', n, from, result) then exit;
  if ALIsNil(n) then noLen := true
  else
  begin
    noLen := false;
    if not ALNextIntArg('VEC-EXTRACT', n, len, result) then exit;
  end;
  result := vec;

  vlen := ALGetVectorLength(vec);
  if noLen then len := vlen;

  if from < 0 then from := vlen+from;
  if (from < 0) or (from >= vlen) then
  begin
    if ALGetAtomType(vec) = DT_VECTOR then result := ALNewVector(0)
    else result := ALNewByteVector(0);
    exit;
  end;
  if from+len > vlen then len := vlen-from;
  if len < 1 then
  begin
    if ALGetAtomType(vec) = DT_VECTOR then result := ALNewVector(0)
    else result := ALNewByteVector(0);
    exit;
  end;

  //WriteLn('from=', from, '; len=', len);
  if ALGetAtomType(vec) = DT_VECTOR then
  begin
    result := ALNewVector(len);
    tl := TList(al_CDR[vec].vi); tl1 := TList(al_CDR[result].vi);
    vlen := 0;
    while len > 0 do
    begin
      tl1[vlen] := tl[from]; tl.Delete(from);
      Inc(vlen); Dec(len);
    end;
  end
  else
  begin
    result := ALNewByteVector(len);
    tv := TLispByteVector(al_CDR[vec].vi); tv1 := TLispByteVector(al_CDR[result].vi);
    c := 0; from0 := from; len0 := len;
    while len0 > 0 do
    begin
      {$R+}
      tv1.vector[c] := tv.vector[from0];
      {$R-}
      Inc(from0); Inc(c); Dec(len0);
    end;

    while from+len < vlen do
    begin
      {$R+}
      tv.vector[from] := tv.vector[from+len]; Inc(from);
      {$R-}
    end;
    SetLength(tv.vector, vlen-len);
  end;
end;

function ALPVecCopy (n: Integer): Integer;
// (vec-copy vector from [len])
// from<0: from the end of the vector
// no len: till the end of the vector
// returns extracted vector (can be empty)
var
  vec, c, from, len, vlen: Integer;
  from0, len0: Integer;
  noLen: Boolean;
  tl, tl1: TList;
  tv, tv1: TLispByteVector;
begin
  if not ALNextAnyVectorArg('VEC-EXTRACT', n, vec, result) then exit;
  if not ALNextIntArg('VEC-EXTRACT', n, from, result) then exit;
  if ALIsNil(n) then noLen := true
  else
  begin
    noLen := false;
    if not ALNextIntArg('VEC-EXTRACT', n, len, result) then exit;
  end;
  result := vec;

  vlen := ALGetVectorLength(vec);
  if noLen then len := vlen;

  if from < 0 then from := vlen+from;
  if (from < 0) or (from >= vlen) then
  begin
    if ALGetAtomType(vec) = DT_VECTOR then result := ALNewVector(0)
    else result := ALNewByteVector(0);
    exit;
  end;
  if from+len > vlen then len := vlen-from;
  if len < 1 then
  begin
    if ALGetAtomType(vec) = DT_VECTOR then result := ALNewVector(0)
    else result := ALNewByteVector(0);
    exit;
  end;

  //WriteLn('from=', from, '; len=', len);
  if ALGetAtomType(vec) = DT_VECTOR then
  begin
    result := ALNewVector(len);
    tl := TList(al_CDR[vec].vi); tl1 := TList(al_CDR[result].vi);
    vlen := 0;
    while len > 0 do
    begin
      tl1[vlen] := tl[from];
      Inc(from); Inc(vlen); Dec(len);
    end;
  end
  else
  begin
    result := ALNewByteVector(len);
    tv := TLispByteVector(al_CDR[vec].vi); tv1 := TLispByteVector(al_CDR[result].vi);
    c := 0; from0 := from; len0 := len;
    while len0 > 0 do
    begin
      {$R+}
      tv1.vector[c] := tv.vector[from0];
      {$R-}
      Inc(from0); Inc(c); Dec(len0);
    end;
  end;
end;

function ALPVecExtractNth (n: Integer): Integer;
// (vec-extract-nth vector index)
// index<0: from the end of the vector
// returns extracted item
var
  vec, index, vlen: Integer;
  tl: TList;
  tv: TLispByteVector;
begin
  if not ALNextAnyVectorArg('VEC-EXTRACT-NTH', n, vec, result) then exit;
  if not ALNextIntArg('VEC-EXTRACT-NTH', n, index, result) then exit;
  result := ALNil();

  vlen := ALGetVectorLength(vec);

  if index < 0 then index := vlen+index;
  if (index < 0) or (index >= vlen) then
  begin
    if ALGetAtomType(vec) = DT_BYTEVECTOR then result := ALNewIntAtom(0);
    exit;
  end;

  //WriteLn('index=', index, '; len=', len);
  if ALGetAtomType(vec) = DT_VECTOR then
  begin
    tl := TList(al_CDR[vec].vi);
    result := Integer(tl[index]);
    tl.Delete(index);
  end
  else
  begin
    tv := TLispByteVector(al_CDR[vec].vi);
    result := ALNewIntAtom(tv.vector[index]);

    while index+1 < vlen do
    begin
      {$R+}
      tv.vector[index] := tv.vector[index+1];
      {$R-}
      Inc(index);
    end;
    SetLength(tv.vector, vlen-1);
  end;
end;

function ALPVecConcat (n: Integer): Integer;
// (vec-concat vector0 vector1)
// concat vector0 & vector1, store result in vector0
// returns vector0
var
  vec0, vec1: Integer;
  f, vlen0, vlen1: Integer;
begin
  if not ALNextAnyVectorArg('VEC-CONCAT', n, vec0, result) then exit;
  if not ALNextAnyVectorArg('VEC-CONCAT', n, vec1, result) then exit;

  if ALGetAtomType(vec0) <> ALGetAtomType(vec1) then
  begin
    result := ALThrowError('VEC-CONCAT: vector types mismatch');
    exit;
  end;

  result := vec0;

  vlen0 := ALGetVectorLength(vec0); vlen1 := ALGetVectorLength(vec1);
  ALSetVectorLength(vec0, vlen0+vlen1);
  for f := 0 to vlen1-1 do ALSetVectorItem(vec0, vlen0+f, ALGetVectorItem(vec1, f));
end;

function ALPVecInsertNth (n: Integer): Integer;
// (vec-insert-nth vector index value)
// index<0: from the end of the vector
// returns vector
var
  f, vec, index, value, vlen: Integer;
  tl: TList;
  tv: TLispByteVector;
begin
  //Write('>>'); ALPrintNoQ(n); WriteLn;
  if not ALNextAnyVectorArg('VEC-INSERT-NTH', n, vec, result) then exit;
  if not ALNextIntArg('VEC-INSERT-NTH', n, index, result) then exit;
  result := vec;

  vlen := ALGetVectorLength(vec);

  if index < 0 then index := vlen+index;
  if (index < 0) or (index > vlen) then exit;

  //WriteLn('index=', index, '; len=', len);
  if ALGetAtomType(vec) = DT_VECTOR then
  begin
    value := al_CAR[n];
    tl := TList(al_CDR[vec].vi);
    tl.Insert(index, Pointer(value));
  end
  else
  begin
    if not ALNextIntArg('VEC-INSERT-NTH', n, value, result) then exit;
    tv := TLispByteVector(al_CDR[vec].vi);
    SetLength(tv.vector, vlen+1);
    f := vlen;
    while f > index do
    begin
      {$R+}
      tv.vector[f] := tv.vector[f-1];
      {$R-}
      Dec(f);
    end;
    tv.vector[index] := value and $FF;
  end;
end;

function ALPVecInsert (n: Integer): Integer;
// (vec-insert vector0 index vector1)
// modifies vector0
// returns vector0
var
  vec0, vec1, index: Integer;
  f, vlen0, vlen1: Integer;
begin
  if not ALNextAnyVectorArg('VEC-INSERT', n, vec0, result) then exit;
  if not ALNextIntArg('VEC-INSERT', n, index, result) then exit;
  if not ALNextAnyVectorArg('VEC-INSERT', n, vec1, result) then exit;

  if ALGetAtomType(vec0) <> ALGetAtomType(vec1) then
  begin
    result := ALThrowError('VEC-INSERT: vector types mismatch');
    exit;
  end;

  result := vec0;

  vlen1 := ALGetVectorLength(vec1); if vlen1 = 0 then exit;
  vlen0 := ALGetVectorLength(vec0);

  if index < 0 then index := vlen0+index;
  if (index < 0) or (index > vlen0) then exit;

  ALSetVectorLength(vec0, vlen0+vlen1);
  f := vlen0+vlen1-1;
  while f >= index+vlen1 do
  begin
    ALSetVectorItem(vec0, f, ALGetVectorItem(vec0, f-vlen1));
    Dec(f);
  end;
  for f := 0 to vlen1-1 do ALSetVectorItem(vec0, index+f, ALGetVectorItem(vec1, f));
end;
