// DarkSpecter ZX Spectum Emulator
// coded by Ketmar // Vampire Avalon (psyc://ketmar.no-ip.org/~Ketmar)
// Understanding is not required. Only obedience.
//
// This program is free software. It comes without any warranty, to
// the extent permitted by applicable law. You can redistribute it
// and/or modify it under the terms of the Do What The Fuck You Want
// To Public License, Version 2, as published by Sam Hocevar. See
// http://sam.zoy.org/wtfpl/COPYING for more details.
//
//!!! ADD LISP API!
// simple debugger
//
{.$DEFINE DEBUGFORTH}
unit dspDbgU;

interface

uses
  Windows, Messages, SysUtils, Classes, lispCon3U, xpackU,
  dspUtilsU,
  z80U, dspEmuVarsU, dspZXDrawU,
  dspUnasmU, zxFloatU, dspEmuMiscU,
  dspMachConstU, dspVScrU, dspLabelsU;


function DbgIsBP (addr: Word): Boolean;
procedure DbgSetBP (addr: Word);
procedure DbgResetBP (addr: Word);
procedure DbgToggleBP (addr: Word);
procedure DbgClearBPs ();

function DbgSaveBPs (fileName: string): Boolean;
function DbgLoadBPs (fileName: string): Boolean;

procedure DoDebugger ();


var
  emuTempBP: Integer = -1; // debugger F8 %-)


implementation


const
  pageH = 20;
  regColors: array [Boolean] of Byte = (15, 7);


var
  bpMap: packed array [0..65535] of Boolean;
  pz80: TZ80CPU;
  unPage: array [0..pageH] of TUnasmInfo;
  oldVScr: packed array [0..239, 0..319] of Byte;

  inDebugger: Boolean = false;


procedure DrawRegs (x, y: Integer); forward;
procedure DrawStack (x, y: Integer); forward;

procedure SaveScreen ();
begin
  Move(dspVScreenRam, oldVScr, SizeOf(oldVScr));
end;

procedure RestoreScreen ();
begin
  Move(oldVScr, dspVScreenRam, SizeOf(dspVScreenRam));
end;

procedure BuildUnPage (addr: Word);
var
  f: Integer;
begin
  for f := 0 to pageH do
  begin
    ZCmdDecode(addr, unpage[f]);
    Inc(addr, unpage[f].len);
  end;
end;

procedure DrawUnPage (wx, wy, cy: Integer);
var
  f, c: Integer;
  ta: Byte;
  s, l: string;
  lf: Boolean;
begin
  for f := 0 to pageH-1 do
  begin
    if f = cy then
    begin
      if unpage[f].addr = emuZ80.regPC then VFillRect(wx, wy, 52*6, 8, 15, 255)
      else VFillRect(wx, wy, 52*6, 8, 7, 255);
      if DbgIsBP(unpage[f].addr) then ta := 2 else ta := 0;
    end
    else
    begin
      if DbgIsBP(unpage[f].addr) then ta := 2 else ta := 7;
      if unpage[f].addr = emuZ80.regPC then
      begin
        Inc(ta, 8);
        VFillRect(wx, wy, 52*6, 8, 1, 255);
      end;
    end;

    VDrawString(wx, wy, IntToHex(unpage[f].addr, 4), ta);
    for c := 0 to unpage[f].len-1 do
      VDrawString(wx+5*6+c*12, wy, IntToHex(unpage[f].dump[c], 2), ta);

    s := unpage[f].cmdName;
    if unpage[f].argC > 0 then
    begin
      while Length(s) < 5 do s := s+' ';
      s := s+unpage[f].args[0];
      for c := 1 to 2 do
        if c < unpage[f].argC then s := s+','+unpage[f].args[c];
    end;
    repeat s := s+' '; until Length(s) > 16; s := s+';';

    l := ZFindCodeLabel(unpage[f].addr, lf);
    if lf then s := s+' '+l+':';
    if (unpage[f].ioFlags and unasmMemReadFlag) <> 0 then
    begin
      l := ZFindDataLabel(unpage[f].memSrcAddr, lf);
      if lf then s := s+' r:'+l+' ';
    end;
    if (unpage[f].ioFlags and unasmMemWriteFlag) <> 0 then
    begin
      l := ZFindDataLabel(unpage[f].memDestAddr, lf);
      if lf then s := s+' w:'+l+' ';
    end;
    if (s <> '') and (s[Length(s)] = ';') then s := TrimRight(Copy(s, 1, Length(s)-1));

    VDrawString(wx+16*6, wy, s, ta);

    Inc(wy, 8);
  end;
end;

procedure DoDebugger ();
var
  wx, wy: Integer;
  msg: TMsg;
  repaint: Boolean;
  curAddr, ad: Word;
  curY: Integer;
  c{, f, n}: Integer;
  addrB: Integer;
  newAddr: Word;
  newTS: Boolean;
  s: string;
  defPau: Boolean;
  escExits: Boolean;
begin
  if not emuSingleStep and (emuTempBP <> Integer(emuZ80.regPC)) and
     not bpMap[emuZ80.regPC] then exit;
  if inDebugger then exit;

  defPau := config.pauseOnDefocus;
  try
    escExits := true;
    emuInDebugger := true;
    config.pauseOnDefocus := false;

    SaveScreen();

    emuTempBP := -1; emuSingleStep := false;
    wx := 0; wy := 4*8+6;

    repaint := true; newAddr := 0;

    BuildUnPage(unpage[0].addr);
    c := pageH-1; while (c >= 0) and (unpage[c].addr <> emuZ80.regPC) do Dec(c);
    if c < 0 then
    begin
      curY := 0; BuildUnPage(emuZ80.regPC);
    end
    else
    begin
      curY := c;
    end;
    addrB := -1; newTS := false; s := '';
    repeat
      BuildUnPage(unpage[0].addr);
      curAddr := unpage[0].addr;
      if repaint or emuWConVisible or (emuWConHeight > 0) then
      begin
        RestoreScreen();
        //EmuRenderScreenNoInv();
        DrawRegs(0, 0);
        DrawStack(32*6+8, 0);

        VFillRect(wx, wy, 52*6+6, pageH*8+6, 0, 0);
        VFillRect(wx+1, wy+1, 52*6+4, pageH*8+4, 255, 7);
        DrawUnPage(wx+3, wy+3, curY);

        if addrB >= 0 then
        begin
          if newTS then
          begin
            VDrawString(wx+3, wy+3, '     ', 15, 9);
            VDrawString(wx+3, wy+3, s, 15, 9);
          end
          else
          begin
            VDrawString(wx+3, wy+3, IntToHex(newAddr, 4), 15, 1);
            VDrawChar(wx+3+addrB*6, wy+3, IntToHex(newAddr, 4)[addrB+1], 15, 11);
          end;
        end;

        if newTS and (addrB < 0) then
        begin
          VDrawString(3+27*6, 3+8*1, '     ', 0, 7);
          VDrawString(3+27*6, 3+8*1, s, 0, 7);
        end;

        DrawTextOutline(300, 2, 'dbg', 15, 0);
        DrawTextOutline(300-6*2, 12, '#'+IntToHex(emuZ80.prevPC, 4), 15, 0);

        if emuWConVisible or (emuWConHeight > 0) then EmuDrawWCon();

        EmuDrawScreen();
        repaint := false;
        escExits := true;
      end;

      if EmuReadKeyMouse(msg) then
      begin
        if (msg.message = WM_QUIT) or (msg.message = WM_CLOSE) or (msg.message = WM_DESTROY) then
        begin
          ConExecute('pause tan');
          PostQuitMessage(0);
          break;
        end;

        if msg.message = WM_PAINT then repaint := true;

        if emuWConVisible or (emuWConHeight > 0) then repaint := true;
        EmuProcessConMessage(msg);
        if emuWConVisible then
        begin
          if (msg.message = WM_KEYDOWN) and ((msg.wParam and $FFFF) = $C0) then emuWConVisible := false
          else EmuProcessWinMessage(msg);
        end
        else if msg.message = WM_KEYUP then
        begin
        end
        else if msg.message = WM_KEYDOWN then
        begin
          //EmuFixShiftsState();
          //emuShiftsState, emuSSAlt
          case msg.wParam and $FFFF of
            $C0: emuWConVisible := true;
            VK_F4:
              begin
                //RestoreScreen();
                //DrawTextOutline(300, 2, 'dbg', 15, 0);
                for c := 0 to 239 do ZXUpdateLine(c, @(emuMemory[emuScreenPage]));
                SaveScreen();
                EmuDrawScreen();
                escExits := false;
                //repaint := true;
              end;
            VK_F5:
              begin
                RestoreScreen();
                DrawTextOutline(300, 2, 'dbg', 15, 0);
                EmuDrawScreen();
                escExits := true;
              end;
            ord('G'):
              if not newTS then
              begin
                if addrB < 0 then begin newAddr := curAddr; addrB := 0; end;
                repaint := true;
              end;
            ord('Z'):
              if not newTS then
              begin
                if addrB < 0 then
                begin
                  newAddr := curAddr; addrB := 0; newTS := true; s := '';
                end;
                repaint := true;
              end;
            ord('T'):
              if not newTS and (addrB = -1) then
              begin
                s := '';
                newTS := true;
                repaint := true;
              end;
            ord('0')..ord('9'), ord('A')..ord('F'):
              if addrB >= 0 then
              begin
                if newTS then
                begin
                  if Length(s) < 5 then s := s+chr(msg.wParam and $FFFF);
                end
                else
                begin
                  Inc(addrB);
                  s := IntToHex(newAddr, 4);
                  s[addrB] := chr(msg.wParam and $FFFF);
                  newAddr := StrToInt('$'+s);
                  if addrB > 3 then addrB := 3;
                end;
                repaint := true;
              end
              else if newTS and ((msg.wParam and $FFFF) < ord('A')) then
              begin
                if Length(s) < 5 then s := s+chr(msg.wParam and $FFFF);
                repaint := true;
              end
              else if (msg.wParam and $FFFF) = ord('C') then
              begin
                DbgClearBPs();
                repaint := true;
              end;
            VK_RETURN:
              if addrB >= 0 then
              begin
                addrB := -1;
                if newTS then
                begin
                  c := StrToIntDef(s, -1);
                  if (c >= 0) and (c < 65536) then begin curAddr := c; curY := 0; end;
                end
                else begin curAddr := newAddr; curY := 0; end;
                BuildUnPage(curAddr);
                newTS := false;
                repaint := true;
              end
              else if newTS and (s <> '') then
              begin
                c := StrToInt(s);
                if c < emuTicksIntr then
                begin
                  emuZ80.tStates := c;
                  newTS := false;
                  repaint := true;
                end;
              end
              else repaint := true;
            VK_BACK:
              if newTS and (s <> '') then
              begin
                Delete(s, Length(s), 1);
                repaint := true;
              end
              else if addrB > 0 then
              begin
                Dec(addrB);
                repaint := true;
              end;
            VK_ESCAPE:
              begin
                if addrB >= 0 then
                begin
                  addrB := -1;
                  repaint := true;
                end
                else
                begin
                  repaint := true;
                  while (GetAsyncKeyState(VK_ESCAPE) and $8000) <> 0 do Sleep(0);
                  if escExits then break;
                end;
              end;
            VK_F1: ;
            VK_F7: begin emuSingleStep := true; break; end;
            VK_F8: begin emuTempBP := (emuZ80.regPC+ZCmdLength(emuZ80.regPC)) and $FFFF; break; end;
            VK_PAUSE: begin ConExecute('pause tan'); repaint := true; {break;} end;
            VK_LEFT:
              begin
                curAddr := (curAddr-1) and $FFFF;
                BuildUnPage(curAddr);
                repaint := true;
              end;
            VK_RIGHT:
              begin
                curAddr := (curAddr+1) and $FFFF;
                BuildUnPage(curAddr);
                repaint := true;
              end;
            VK_UP:
              begin
                if curY > 0 then Dec(curY)
                else
                begin
                  ad := (curAddr-5) and $FFFF;
                  while ad <> curAddr do
                  begin
                    if ((ad+ZCmdLength(ad)) and $FFFF) = curAddr then break;
                    ad := (ad+1) and $FFFF;
                  end;
                  if ad = curAddr then curAddr := (curAddr-1) and $FFFF
                  else curAddr := ad;
                  BuildUnPage(curAddr);
                end;
                repaint := true;
              end;
            VK_PRIOR:
              begin
                for c := 2 to pageH do
                begin
                  ad := (curAddr-5) and $FFFF;
                  while ad <> curAddr do
                  begin
                    if ((ad+ZCmdLength(ad)) and $FFFF) = curAddr then break;
                    ad := (ad+1) and $FFFF;
                  end;
                  if ad = curAddr then curAddr := (curAddr-1) and $FFFF
                  else curAddr := ad;
                end;
                BuildUnPage(curAddr);
                repaint := true;
              end;
            VK_DOWN:
              begin
                if curY < pageH-1 then Inc(curY)
                else BuildUnPage(unpage[1].addr);
                repaint := true;
              end;
            VK_NEXT:
              begin
                BuildUnPage(unpage[pageH-1].addr);
                repaint := true;
              end;
            VK_SPACE:
              begin
                DbgToggleBP(unpage[curY].addr);
                repaint := true;
              end;
            ord('S'):
              begin
                DbgSaveBPs('dspdef');
                repaint := true;
              end;
            ord('L'):
              begin
                DbgLoadBPs('dspdef');
                repaint := true;
              end;
            //else EmuProcessWinMessage(msg);
          end;
        end
        else EmuProcessWinMessage(msg);
      end
      else Sleep(10);
    until false;
    Move(emuZ80^, pz80, SizeOf(pz80));
    RestoreScreen();
    if repaint then EmuDrawScreen();
    emuInDebugger := false;
    EmuResetTimers();
    WaitKeybRelease();
  finally
    config.pauseOnDefocus := defPau;
    inDebugger := false;
  end;
end;


procedure DrawStack (x, y: Integer);
var
  f: Integer;
  a: Word;
begin
  VFillRect(x, y, 10*6+6, 4*8+6, 0, 0);
  VFillRect(x+1, y+1, 10*6+4, 4*8+4, 255, 7);

  Inc(x, 3); Inc(y, 3);
  a := (emuZ80.regSP-2) and $FFFF;
  for f := 0 to 3 do
  begin
    VDrawString(x, y+f*8, IntToHex(Z80PeekBInternal(emuZ80, (a+1) and $FFFF), 2), 15, 255);
    VDrawString(x+2*6, y+f*8, IntToHex(Z80PeekBInternal(emuZ80, a), 2), 15, 255);

    VDrawString(x+6*6, y+f*8, IntToHex(Z80PeekBInternal(emuZ80, (a+9) and $FFFF), 2), 15, 255);
    VDrawString(x+8*6, y+f*8, IntToHex(Z80PeekBInternal(emuZ80, (a+8) and $FFFF), 2), 15, 255);
    Inc(a, 2);
  end;
  a := (emuZ80.regSP-2) and $FFFF;
  VDrawString(x, y, IntToHex(Z80PeekBInternal(emuZ80, (a+1) and $FFFF), 2), 7, 255);
  VDrawString(x+2*6, y, IntToHex(Z80PeekBInternal(emuZ80, a), 2), 7, 255);
end;

procedure DrawHexB (x, y: Integer; b: Byte; unchanged: Boolean);
begin
  VDrawString(x, y, IntToHex(b, 2), regColors[unchanged]);
end;

procedure DrawHexW (x, y: Integer; w: Word; changed: Boolean);
begin
  VDrawString(x, y, IntToHex(w, 4), regColors[changed]);
end;


procedure DrawRegs (x, y: Integer);
const
  zo: array [Boolean] of Char = ('0', '1');
  flg: array [Boolean, 0..7] of Char = ('sz.h.pnc', 'SZ5H3PNC');
var
  b0, b1: Byte;
  s: string;
  f: Integer;
begin
  VFillRect(x, y, 32*6+6, 4*8+6, 0, 0);
  VFillRect(x+1, y+1, 32*6+4, 4*8+4, 255, 7);

  VDrawString(x+3, y+3+8*0, 'af:     af''     sp:     ir:', 7);
  VDrawString(x+3, y+3+8*1, 'bc:     bc''     pc:     t:', 7);
  VDrawString(x+3, y+3+8*2, 'de:     de''     ix:     im ,i:', 7);
  VDrawString(x+3, y+3+8*3, 'hl:     hl''     iy:', 7);

  DrawHexW(x+3+3*6, y+3+8*0, Z80GetAF(emuZ80), Z80GetAF(emuZ80)=Z80GetAF(@pz80));
  DrawHexW(x+3+3*6, y+3+8*1, Z80GetBC(emuZ80), Z80GetBC(emuZ80)=Z80GetBC(@pz80));
  DrawHexW(x+3+3*6, y+3+8*2, emuZ80.regDE, emuZ80.regDE=pz80.regDE);
  DrawHexW(x+3+3*6, y+3+8*3, emuZ80.regHL, emuZ80.regHL=pz80.regHL);

  DrawHexW(x+3+11*6, y+3+8*0, emuZ80.regAFx, emuZ80.regAFx=pz80.regAFx);
  DrawHexW(x+3+11*6, y+3+8*1, emuZ80.regBCx, emuZ80.regBCx=pz80.regBCx);
  DrawHexW(x+3+11*6, y+3+8*2, emuZ80.regDEx, emuZ80.regDEx=pz80.regDEx);
  DrawHexW(x+3+11*6, y+3+8*3, emuZ80.regHLx, emuZ80.regHLx=pz80.regHLx);

  DrawHexW(x+3+19*6, y+3+8*0, emuZ80.regSP, emuZ80.regSP=pz80.regSP);
  DrawHexW(x+3+19*6, y+3+8*1, emuZ80.regPC, emuZ80.regPC=pz80.regPC);
  DrawHexW(x+3+19*6, y+3+8*2, emuZ80.regIX, emuZ80.regIX=pz80.regIX);
  DrawHexW(x+3+19*6, y+3+8*3, emuZ80.regIY, emuZ80.regIY=pz80.regIY);

  b0 := (emuZ80.regR and $80) or (emuZ80.regRTemp and $7F);
  b1 := (pz80.regR and $80) or (pz80.regRTemp and $7F);
  DrawHexB(x+3+28*6, y+3+8*0, emuZ80.regI, emuZ80.regI=pz80.regI);
  DrawHexB(x+3+30*6, y+3+8*0, b0, b0=b1);
  Str(emuZ80.tStates:5, s);
  VDrawString(x+3+27*6, y+3+8*1, s, regColors[true]);

  VDrawChar(x+3+26*6, y+3+8*2, chr(emuZ80.im+ord('0')), regColors[emuZ80.im=pz80.im]);
  VDrawChar(x+3+30*6, y+3+8*2, zo[emuZ80.iff1], regColors[emuZ80.iff1=pz80.iff1]);
  VDrawChar(x+3+31*6, y+3+8*2, zo[emuZ80.iff2], regColors[emuZ80.iff2=pz80.iff2]);

  b0 := Z80GetF(emuZ80); b1 := Z80GetF(@pz80);
  for f := 0 to 7 do
  begin
    VDrawChar(x+3+(24+f)*6, y+3+8*3, flg[(b0 and $80) <> 0, f], regColors[(b0 and $80)=(b1 and $80)]);
    b0 := (b0 and $7F) shl 1; b1 := (b1 and $7F) shl 1;
  end;
end;


function DbgSaveBPs (fileName: string): Boolean;
var
  fo: TFileStream;
  buf: packed array of Byte;
  r: Integer;
  sign: packed array [0..3] of Char;
begin
  result := false;
  if fileName = '' then exit;
  if ExtractFileExt(fileName) = '' then fileName := fileName+'.dbp';
  try
    fo := TFileStream.Create(fileName, fmCreate);
  except
    try
      fo := TFileStream.Create(fileName, fmOpenWrite or fmShareDenyWrite);
    except exit;
    end;
  end;
  try
    try
      sign := 'DBP0';
      SetLength(buf, 65536);
      r := XPackBuffer(bpMap[0], 65536, buf[0], 65535);
      if r = -1 then
      begin
        r := 65536; Move(bpMap[0], buf[0], 65536);
      end;
      fo.WriteBuffer(sign[0], 4);
      fo.WriteBuffer(r, 2);
      fo.WriteBuffer(buf[0], r);
    finally
      fo.Free();
    end;
    result := true;
  except
    DeleteFile(fileName);
    exit;
  end;
end;

function DbgLoadBPs (fileName: string): Boolean;
var
  fi: TFileStream;
  buf: packed array of Byte;
  r: Integer;
  sign: packed array [0..3] of Char;
begin
  result := false;
  if fileName = '' then exit;
  if ExtractFileExt(fileName) = '' then fileName := fileName+'.dbp';
  try
    fi := TFileStream.Create(fileName, fmOpenRead or fmShareDenyWrite);
  except exit;
  end;
  try
    try
      fi.ReadBuffer(sign[0], 4);
      if sign <> 'DBP0' then exit;
      r := 0; fi.ReadBuffer(r, 2); if r = 0 then r := 65536;
      SetLength(buf, 65536);
      fi.ReadBuffer(buf[0], r);
      if r = 65536 then
      begin
        Move(buf[0], bpMap[0], 65536);
        result := true;
      end
      else
      begin
        r := XUnpackBuffer(buf[0], r, bpMap[0], 65536);
        if r <> 65536 then DbgClearBPs() else result := true;
      end;
    finally
      fi.Free();
    end;
    result := true;
  except exit;
  end;
end;

procedure DbgClearBPs ();
begin
  FillChar(bpMap, SizeOf(bpMap), 0);
end;

function DbgIsBP (addr: Word): Boolean;
begin
  result := bpMap[addr];
end;

procedure DbgSetBP (addr: Word);
begin
  bpMap[addr] := true;
end;

procedure DbgResetBP (addr: Word);
begin
  bpMap[addr] := false;
end;

procedure DbgToggleBP (addr: Word);
begin
  bpMap[addr] := not bpMap[addr];
end;


initialization
  DbgClearBPs();
  FillChar(pz80, SizeOf(pz80), 0);
  unpage[0].addr := 0;
finalization
end.
