// URAL scipting language v0.2.0a
// coded by Ketmar // Vampire Avalon (psyc://ketmar.no-ip.org/~Ketmar)
// Understanding is not required. Only obedience.
//
// This program is free software. It comes without any warranty, to
// the extent permitted by applicable law. You can redistribute it
// and/or modify it under the terms of the Do What The Fuck You Want
// To Public License, Version 2, as published by Sam Hocevar. See
// http://sam.zoy.org/wtfpl/COPYING for more details.
//
// macro language interpreter
//
// basic-like language (weak typed)
// the language is fast enough to be usable %-)
// the language is case-insensitive
//
{$H-,R-,S-,Q-,W-,A8,Z1} // for speed
{.$DEFINE XCOMPILE}
unit uralU;

interface


const
  TK_EOP = Pointer(1); // end of operator. corresponding string is line no or ''
  TK_KEYWORD = nil; // +,-, etc: keywords!
  TK_ID = Pointer(2);
  TK_OTHER = Pointer(3);

  KW_SUB      = #1;
  KW_SUBEND   = #2;
  KW_IF       = #3;
  KW_THEN     = #4;
  KW_ELSE     = #5;
  KW_ENDIF    = #6;
  KW_ELSIF    = #7;
  KW_GOELSE   = #8;
  KW_BREAKIF  = #9;
  KW_WHILE    = #10;
  KW_WEND     = #11;
  KW_AGAIN    = #12;
  KW_BREAK    = #13;
  KW_RETURN   = #14;

  KW_CONCAT   = #15; // ||
  KW_LESSEQU  = #16; // <=
  KW_GREATEQU = #17; // >=
  KW_NOTEQU   = #18; // <>
  KW_LOGAND   = #19; // &
  KW_LOGOR    = #20; // |
  KW_LOGXOR   = #21; // ^
  KW_PLUS     = #22; // +
  KW_MINUS    = #23; // -
  KW_MUL      = #24; // *
  KW_DIV      = #25; // /
  KW_IDIV     = #26; // \
  KW_MOD      = #27; // MOD
  KW_BITAND   = #28; // AND
  KW_BITOR    = #29; // OR
  KW_BITXOR   = #30; // XOR
  KW_LESS     = #31; // <
  KW_GREAT    = #32; // >
  KW_LOGNOT   = #33; // !
  KW_BITNOT   = #34; // ~
  KW_EQU      = #35; // =


type
  PURALSSArray = ^TURALSSArray;
  TURALSSArray = packed array [0..65535] of ShortString;
  PURALOArray = ^TURALOArray;
  TURALOArray = packed array [0..65535] of TObject;

  // list of shortstrings
  TURALStrList = class
  protected
    fItems: PURALSSArray;
    fObjects: PURALOArray;
    fCapacity: Integer;
    fCount: Integer;
    fSorted: Boolean;

    function GetCount (): Integer;
    function GetItems (index: Integer): ShortString;
    procedure SetItems (index: Integer; const v: ShortString);
    function GetObjects (index: Integer): TObject;
    procedure SetObjects (index: Integer; v: TObject);

    procedure Grow (n: Integer);

  public
    constructor Create (pSorted: Boolean = false);
    destructor Destroy (); override;

    procedure Insert (index: Integer; const s: ShortString);
      // do not insert strings in sorted list!
    function Add (const s: ShortString): Integer;
    function AddObject (const s: ShortString; o: TObject): Integer;
    procedure Delete (index: Integer);
    procedure Clear ();

    function Find (const s: ShortString; out index: Integer): Boolean;
    function IndexOf (const s: ShortString): Integer;

    property Count: Integer read GetCount;
    property Items [index: Integer]: ShortString read GetItems write SetItems; default;
    //property ItemAddr [index: Integer]: PChar read GetItemAddr;
    property Objects [index: Integer]: TObject read GetObjects write SetObjects;
  end;

  TURALInterpreter = class;

  // user-defined function
  TURALFunction = class
  public
    function ExecuteEx (intr: TURALInterpreter; args: TURALStrList; const fName: ShortString): ShortString; virtual;
      // this simply calls Execute()
    function Execute (intr: TURALInterpreter; args: TURALStrList): ShortString; virtual;
      // do not change args!
  end;

  TURALFuncMath = function (const a0, a1: ShortString): ShortString of object;

  TURALInterpreter = class
  protected
    fFuncList: TURALStrList; // sorted!
      // Items[]: name
      // Objects[]: TURALStrList with tokens
      // tokenized list is:
      //   Items[]: token values or keyword ids
      //   Objects[]: token types
      // first item: execution counter (see history.txt)
    fExtFuncs: TURALStrList; // sorted!
      // Items[]: name
      // Objects[]: TURALFunction
    fStdMath: array [#15..#35] of TURALFuncMath;
      // keep the bounds in sync with KW_xxx!

    fError: Boolean; // last Execute() failed?
    fErrorMsg: ShortString; // last error message
    fErrorFunc: ShortString;
    fErrorLine: Integer;

    fVarPool: array of record name, value: ShortString; end;
    fVarPoolUsed: Integer;
    fVarPoolMax: Integer; // very internal thing %-)

    fCurPoolTop: Integer;
    fCurArgs: TURALStrList; // current function arguments (for argc & argv)
    fCurIP: Integer;
    fCurFuncName: ShortString;
    fCurFunc: TURALStrList;

    fInstructionLimit: Integer;
    fInstructionCount: Integer;

    fInDoExecute: Boolean;

    function DoMathPlus (const a0, a1: ShortString): ShortString;
    function DoMathMinus (const a0, a1: ShortString): ShortString;
    function DoMathMul (const a0, a1: ShortString): ShortString;
    function DoMathDiv (const a0, a1: ShortString): ShortString;
    function DoMathIDiv (const a0, a1: ShortString): ShortString;
    function DoMathMod (const a0, a1: ShortString): ShortString;
    function DoMathConcat (const a0, a1: ShortString): ShortString;
    function DoMathLess (const a0, a1: ShortString): ShortString;
    function DoMathGreat (const a0, a1: ShortString): ShortString;
    function DoMathEqu (const a0, a1: ShortString): ShortString;
    function DoMathGreatEqu (const a0, a1: ShortString): ShortString;
    function DoMathLessEqu (const a0, a1: ShortString): ShortString;
    function DoMathNotEqu (const a0, a1: ShortString): ShortString;
    function DoMathLogAnd (const a0, a1: ShortString): ShortString;
    function DoMathLogOr (const a0, a1: ShortString): ShortString;
    function DoMathLogXor (const a0, a1: ShortString): ShortString;
    function DoMathLogNot (const a0, a1: ShortString): ShortString;
    function DoMathBitNot (const a0, a1: ShortString): ShortString;
    function DoMathBitAnd (const a0, a1: ShortString): ShortString;
    function DoMathBitOr (const a0, a1: ShortString): ShortString;
    function DoMathBitXor (const a0, a1: ShortString): ShortString;

    procedure RegisterStandardFunctions (); virtual;

    function CollectTrain (const leaderName: ShortString): ShortString;
    procedure ParseFunctionArgs (al: TURALStrList);
    function DoFunctionCall (const funcname: ShortString): ShortString;
    function Expression (): ShortString; // evaluate expression from fCurIP

    function AnsiStringToStrList (const st: AnsiString): TURALStrList;
    function ShortStringToStrList (const st: ShortString): TURALStrList;

  public
    constructor Create ();
    destructor Destroy (); override;

    procedure Error (const msg: ShortString);

    function GetLocalVar (const pName: ShortString): ShortString; virtual;
      // name MUST be in upper case!
    procedure SetLocalVar (const pName, value: ShortString); virtual;
      // name MUST be in upper case!

    function GetLocalVarCI (pName: ShortString): ShortString;
      // name will be uppercased
    procedure SetLocalVarCI (pName: ShortString; const value: ShortString);
      // name will be uppercased

    function IsGlobalExists (const pName: ShortString): Boolean; virtual;
      // name MUST be in upper case!
    function GetGlobalVar (const pName: ShortString): ShortString; virtual;
      // name MUST be in upper case!
    procedure SetGlobalVar (const pName, value: ShortString); virtual;
      // name MUST be in upper case!

    function IsGlobalExistsCI (pName: ShortString): Boolean;
      // name MUST be in upper case!
    function GetGlobalVarCI (pName: ShortString): ShortString;
      // name MUST be in upper case!
    procedure SetGlobalVarCI (pName: ShortString; const value: ShortString);
      // name MUST be in upper case!

    function Tokenize (pFunc: TURALStrList): TURALStrList;

    function PrecompileFunction (pFunc: TURALStrList; out cfn: TURALStrList;
      tokenized: Boolean = false): ShortString;
      // returns the name of the function
      // if cfn = nil, there was errors
      // if tokenized, just build the header
      // anyway, cfn is a new list

    procedure Clear (); // does not clear external funcs
    procedure AddFunction (pFunc: TURALStrList; tokenized: Boolean = false); overload;
      // replaces old function (if there was any)
      // ignores nils
    procedure AddFunction (const pFunc: ShortString); overload;
    procedure AddFunctionLong (const pFunc: AnsiString);
    function HasFunction (const pName: ShortString): Boolean;

    function AddTextSL (const txt: TURALStrList): ShortString;
      // load the whole text file to string and pass it here
      // interpreter will parse it and add the corresponding functions
      // returns error message or ''
    function AddText (const txt: AnsiString): ShortString;
      // load the whole text file to string and pass it here
      // interpreter will parse it and add the corresponding functions
      // returns error message or ''

    function FunctionListSL (const txt: TURALStrList): TURALStrList;
      // returns the list of the functions in the source
    function FunctionList (const txt: AnsiString): TURALStrList;
      // returns the list of the functions in the source

    procedure DeleteFunction (const pName: ShortString);

    procedure ClearExternalFunc ();
    procedure AddExternalFunc (const pName: ShortString; pUFunc: TURALFunction);
      // ignores nils
    procedure DeleteExternalFunc (const pName: ShortString);
    function FindExternalFunc (const pName: ShortString): TURALFunction;

    function DoExecute (pFuncName: ShortString; args: TURALStrList;
      thisSub: TURALStrList = nil): ShortString;
      // call DoExecute() instead Execute() from TURALFunction.
      // NEVER set thisSub to anything than NIL!

    function Execute (pFuncName: ShortString; const args: array of ShortString): ShortString; overload;
    function Execute (pFuncName: ShortString; args: TURALStrList): ShortString; overload;
      // check Error property to see if there was errors
      // NEVER call Execute() from TURALFunction! use DoExecute() instead.

    function EvalSub (pSubBody: TURALStrList; pArgs: array of ShortString): ShortString; overload;
    function EvalSub (pSubBody, pArgs: TURALStrList): ShortString; overload;
    function EvalSub (const pSubBody: ShortString; pArgs: array of ShortString): ShortString; overload;
    function EvalSub (const pSubBody: ShortString; pArgs: TURALStrList): ShortString; overload;
    function EvalSubLong (const pSubBody: AnsiString; pArgs: array of ShortString): ShortString; overload;
    function EvalSubLong (const pSubBody: AnsiString; pArgs: TURALStrList): ShortString; overload;

    property InstructionLimit: Integer read fInstructionLimit write fInstructionLimit;
      // Execute() executes not more than InstructionLimit operators
    property ErrorFlag: Boolean read fError write fError;
    property ErrorMsg: ShortString read fErrorMsg;
    property ErrorFunc: ShortString read fErrorFunc;
    property ErrorLine: Integer read fErrorLine;

    property CurArgs: TURALStrList read fCurArgs;
      // current executing function arguments
    property CurIP: Integer read fCurIP write fCurIP;
      // be carful with this! %-)
    property CurFuncName: ShortString read fCurFuncName;
    property CurFunc: TURALStrList read fCurFunc;
      // current function, tokenized. CurIP is the index in this list.
  end;


const
  URTT_STRING = 0;
  URTT_INT = 1;
  URTT_FLOAT = 2;


procedure URALFreeAndNil (var obj);

function URALStr2Int (const s: ShortString; out error: Boolean): Integer;
function URALStr2Float (const s: ShortString; out error: Boolean): Double;
function URALInt2Str (i: Integer): ShortString;
function URALFloat2Str (v: Double): ShortString;

function URALIsValidId (const s: ShortString): Boolean;
function URALTokenType (const s: ShortString): Integer; // returns URTT_xxx

function URALBool2Str (b: Boolean): ShortString;
function URALIsTrue (const s: ShortString): Boolean;

procedure URALUpStr (var s: ShortString);

procedure URALTrim (var s: ShortString);


// math functions (to avoid Math unit)
function URALTan (const x: Extended): Extended;

function URALArcTan2 (const y, x: Extended): Extended;
function URALArcCos (const x: Extended): Extended;
  // IN: |x| <= 1  OUT: [0..PI] radians
function URALArcSin (const x: Extended): Extended;
  // IN: |x| <= 1  OUT: [-PI/2..PI/2] radians

function URALRadToDeg (const radians: Extended): Extended;
  // degrees := radians * 180 / PI
function URALDegToRad (const degrees: Extended): Extended;
  // radians := degrees * PI / 180

function CompareSStr (const s0, s1: ShortString): Integer; assembler; register;

function SplitStringToList (const txt: AnsiString): TURALStrList;

function LoadTextFileToList (const fName: ShortString): TURALStrList;
// load the whole text file
// returns NIL if there's no such file


implementation


const
  LIST_GROW_DELTA = 64;


function CompareSStr (const s0, s1: ShortString): Integer; assembler; register;
asm
  xor    ecx,ecx
  cmp    [eax],cl
  jne    @@docmp
@@s0is0:
  xor    eax,eax
  cmp    cl,[edx]
  adc    al,cl
  ret
@@docmp:
  cmp    [edx],cl
  jne    @@realcmp
  xor    eax,eax
  dec    eax
  ret
@@realcmp:
  push   esi
  push   edi
  mov    esi,eax  // ESI=ptr to s0
  mov    edi,edx  // EDI=ptr to s1
  mov    al,[esi] // length of s0
  mov    dl,[edi] // length of s1
@@again:
  inc    esi
  inc    edi
  mov    cl,[esi]
  sub    cl,[edi]
  jnz    @@done
  dec    al
  jz     @@endofs0
  dec    dl
  jnz    @@again
  inc    cl
  jmp    @@done
@@endofs0:
  dec    dl
  jz     @@done
  dec    cl
@@done:
  movsx  eax,cl
@@quit:
  pop    edi
  pop    esi
end;


{ TURALFunction }
function TURALFunction.ExecuteEx (intr: TURALInterpreter; args: TURALStrList; const fName: ShortString): ShortString;
begin
  result := Execute(intr, args);
end;

function TURALFunction.Execute (intr: TURALInterpreter; args: TURALStrList): ShortString;
begin
  result := '';
end;


const
  MAX_PRIO = 8;
  PRIO_RIGHT = 8;
  MAX_OP_AT_LEVEL = 5;
  // priority level 1 -- opertiors line "." and "[]", function calls
  // priority level 2 -- unary opertiors like "!" and "~"
  // priority level 7 -- always right-associative! (PRIO_RIGHT)
  MATH_OPZ: array [2..MAX_PRIO, 0..MAX_OP_AT_LEVEL] of Char = (
    // priority level 2
    (KW_BITAND, #0, #0, #0, #0, #0),
    // priority level 3
    (KW_BITOR, KW_BITXOR, #0, #0, #0, #0),
    // priority level 4
    (KW_MUL, KW_DIV, KW_IDIV, KW_MOD, #0, #0),
    // priority level 5
    (KW_PLUS, KW_CONCAT, KW_MINUS, #0, #0, #0),
    // priority level 6
    (KW_LOGAND, #0, #0, #0, #0, #0),
    (KW_LOGOR, KW_LOGXOR, #0, #0, #0, #0),
    // priority level 7
    (KW_LESS, KW_GREAT, KW_EQU, KW_LESSEQU, KW_GREATEQU, KW_NOTEQU)
  );


{ TURALStrList }
constructor TURALStrList.Create (pSorted: Boolean = false);
begin
  inherited Create();
  fCount := 0; fSorted := pSorted; fCapacity := 0;
  fItems := nil; fObjects := nil;
end;

destructor TURALStrList.Destroy ();
begin
  Clear(); fCapacity := 0; ReallocMem(fItems, 0); ReallocMem(fObjects, 0);
  inherited;
end;

procedure TURALStrList.Clear ();
begin
  fCount := 0;
end;

function TURALStrList.GetCount (): Integer;
begin
  result := fCount;
end;

function TURALStrList.GetItems (index: Integer): ShortString;
begin
  if (index < 0) or (index >= fCount) then result[0] := #0
  else result := fItems^[index];
end;

procedure TURALStrList.SetItems (index: Integer; const v: ShortString);
begin
  if (index >= 0) and (index < fCount) then fItems^[index] := v;
end;

function TURALStrList.GetObjects (index: Integer): TObject;
begin
  if (index < 0) or (index >= fCount) then result := nil
  else if fObjects = nil then result := nil
  else result := fObjects^[index];
end;

procedure TURALStrList.SetObjects (index: Integer; v: TObject);
var
  f: Integer;
begin
  if (index < 0) or (index >= fCount) then exit;
  if fObjects = nil then
  begin
    if v = nil then exit;
    ReallocMem(fObjects, fCapacity*SizeOf(TObject));
    f := fCount-1; while f >= 0 do begin fObjects^[f] := nil; Dec(f); end;
  end;
  fObjects^[index] := v;
end;

function TURALStrList.IndexOf (const s: ShortString): Integer;
begin
  if fSorted then
  begin
    if not Find(s, result) then result := -1;
  end
  else
  begin
    result := fCount-1;
    while (result >= 0) and (fItems^[result] <> s) do Dec(result);
  end;
end;

function TURALStrList.Find (const s: ShortString; out index: Integer): Boolean;
var
  l, h, i, c: Integer;
begin
  if not fSorted then
  begin
    i := IndexOf(s);
    result := (i <> -1);
    if not result then index := fCount else index := i;
    exit;
  end;
  // binary search
  result := false;
  l := 0; h := fCount-1;
  while l <= h do
  begin
    i := (l+h) shr 1;
    c := CompareSStr(fItems^[i], s);
    if c < 0 then l := i+1
    else
    begin
      h := i-1;
      if c = 0 then
      begin
        result := true;
        index := i;
        exit;
      end;
    end;
  end;
  index := l;
end;

procedure TURALStrList.Grow (n: Integer);
begin
  if n < 1 then exit;
  ReallocMem(fItems, (fCapacity+n)*SizeOf(ShortString));
  if fObjects <> nil then
    ReallocMem(fObjects, (fCapacity+n)*SizeOf(TObject));
  Inc(fCapacity, n);
end;

function TURALStrList.Add (const s: ShortString): Integer;
begin
  if not fSorted then
  begin
    result := fCount;
    while result >= fCapacity do Grow(LIST_GROW_DELTA);
    fItems^[result] := s; if fObjects <> nil then fObjects^[result] := nil;
    Inc(fCount);
  end
  else
  begin
    Find(s, result);
    Insert(result, s);
  end;
end;

function TURALStrList.AddObject (const s: ShortString; o: TObject): Integer;
begin
  result := Add(s); if o = nil then exit;
  SetObjects(result, o);
end;

procedure TURALStrList.Insert (index: Integer; const s: ShortString);
var
  c: Integer;
begin
  while fCount >= fCapacity do Grow(LIST_GROW_DELTA);
  if index < 0 then index := 0;
  if index >= fCount then
  begin
    index := fCount;
    fItems^[index] := s; if fObjects <> nil then fObjects^[index] := nil;
    Inc(fCount);
    exit;
  end;
  // make room
  if fObjects <> nil then
  begin
    for c := fCount downto index+1 do
    begin
      fItems^[c] := fItems^[c-1];
      fObjects^[c] := fObjects^[c-1];
    end;
    fObjects^[index] := nil;
  end
  else
  begin
    for c := fCount downto index+1 do fItems^[c] := fItems^[c-1];
  end;
  fItems^[index] := s;
  Inc(fCount);
end;

procedure TURALStrList.Delete (index: Integer);
var
  f: Integer;
begin
  if (index < 0) or (index >= fCount) then exit;
  if fObjects <> nil then
  begin
    for f := index+1 to fCount-1 do
    begin
      fItems^[f-1] := fItems^[f];
      fObjects^[f-1] := fObjects^[f];
    end;
  end
  else for f := index+1 to fCount-1 do fItems^[f-1] := fItems^[f];
  Dec(fCount);
end;


{ utilities }
procedure URALFreeAndNil (var obj);
var
  temp: TObject;
begin
  Temp := TObject(obj);
  if temp <> nil then
  begin
    Pointer(obj) := nil;
    temp.Free();
  end;
end;

function URALIsTrue (const s: ShortString): Boolean;
begin
  result := (s <> '') and (s <> '0') and (s <> '0.0');
end;

function URALBool2Str (b: Boolean): ShortString;
begin
  if b then result := '-1' else result := '0';
end;

function URALIsValidId (const s: ShortString): Boolean;
var
  f: Integer;
begin
  result := false; if Length(s) = 0 then exit;
  if not (s[1] in ['_','A'..'Z','a'..'z']) then exit;
  for f := 2 to Length(s) do
    if not (s[1] in ['0'..'9','_','A'..'Z','a'..'z']) then exit;
  result := true;
end;

function URALTokenType (const s: ShortString): Integer;
var
  f: Integer;
begin
  result := URTT_STRING; if Length(s) = 0 then exit;
  f := 1;
  if s[f] in ['+','-'] then
  begin
    Inc(f); if f > Length(s) then exit;
  end;

  if s[f] in ['0'..'9'] then
  begin
    while (f <= Length(s)) and (s[f] in ['0'..'9']) do Inc(f);
    if f > Length(s) then begin result := URTT_INT; exit; end;
    if s[f] <> '.' then exit;
    Inc(f);
    while (f <= Length(s)) and (s[f] in ['0'..'9']) do Inc(f);
    if f > Length(s) then result := URTT_FLOAT;
  end
  else
  begin
    Inc(f); if f > Length(s) then exit;
    case s[f-1] of
      '%':
        begin
          while (f <= Length(s)) and (s[f] in ['0'..'1']) do Inc(f);
          if f > Length(s) then result := URTT_INT;
        end;
      '@':
        begin
          while (f <= Length(s)) and (s[f] in ['0'..'7']) do Inc(f);
          if f > Length(s) then result := URTT_INT;
        end;
      '#','$':
        begin
          while (f <= Length(s)) and (s[f] in ['0'..'9','A'..'F','a'..'f']) do Inc(f);
          if f > Length(s) then result := URTT_INT;
        end;
    end;
  end;
end;

function URALStr2Int (const s: ShortString; out error: Boolean): Integer;
// s must be valid integer!
var
  neg: Boolean;
  f, len, base: Integer;
  ch: Char;
begin
  error := true;
  result := 0;
  len := Length(s); if len = 0 then exit;
  neg := false; f := 1;
  if s[f] = '-' then begin neg := true; Inc(f); end
  else if s[f] = '+' then Inc(f);
  case s[f] of
    '%': base := 2;
    '@': base := 8;
    '#','$': base := 16;
    else base := 10;
  end;
  if base <> 10 then
  begin
    Inc(f); if f > len then exit;
  end;
  while f <= len do
  begin
    ch := UpCase(s[f]); Inc(f);
    case base of
      2: if not (ch in ['0'..'1']) then exit;
      8: if not (ch in ['0'..'7']) then exit;
      10: if not (ch in ['0'..'9']) then exit;
      16:
        begin
          if not (ch in ['0'..'9','A'..'F']) then exit;
          if ch > '9' then Dec(ch, 7);
        end;
    end;
    result := result*base+ord(ch)-ord('0');
  end;
  if neg then result := -result;
  error := false;
end;

function URALStr2Float (const s: ShortString; out error: Boolean): Double;
// s must be valid integer!
var
  e: Integer;
begin
  Val(s, result, e); if e <> 0 then result := 0.0;
  error := (e <> 0);
end;

function URALInt2Str (i: Integer): ShortString;
begin
  Str(i, result);
end;

function URALFloat2Str (v: Double): ShortString;
var
  f, e: Integer;
begin
  Str(v:36:18, result);
  // trim trailing fractional zeroes
  f := Byte(result[0]);
  while (f > 0) and (result[f] = '0') do Dec(f);
  if f = 0 then
  begin
    Inc(result[0]); result[Byte(result[0])] := '.';
    Inc(result[0]); result[Byte(result[0])] := '0';
  end
  else
  begin
    if result[f] = '.' then Inc(f);
    result[0] := Char(f);
  end;
  f := 1; e := Byte(result[0]);
  while (f <= e) and (result[f] = ' ') do Inc(f);
  if f > 1 then Delete(result, 1, f-1);
end;

procedure URALTrim (var s: ShortString);
var
  b, e: Integer;
begin
  e := Length(s);
  while (e > 0) and (s[e] <= ' ') do Dec(e);
  s[0] := chr(e);
  if e = 0 then exit;
  b := 1; while (b <= e) and (s[b] <= ' ') do Inc(b);
  if b > 1 then Delete(s, 1, b-1);
end;

procedure URALUpStr (var s: ShortString);
var
  f: Integer;
begin
  for f := 1 to Length(s) do s[f] := UpCase(s[f]);
end;


function URALTan (const x: Extended): Extended;
//  tan := sin(x)/cos(x)
asm
  fld    x
  fptan
  fstp   st(0)  // fptan pushes 1.0 after result
  fwait
end;

function URALArcCos (const x: Extended): Extended;
begin
  result := URALArcTan2(sqrt(1-x*x), x);
end;

function URALArcSin (const x: Extended): Extended;
begin
  result := URALArcTan2(x, sqrt(1-x*x))
end;

function URALArcTan2 (const y, x: Extended): Extended;
asm
  fld     y
  fld     x
  fpatan
  fwait
end;

function URALDegToRad (const degrees: Extended): Extended;
begin
  result := degrees*(PI/180);
end;

function URALRadToDeg (const radians: Extended): Extended;
begin
  result := radians*(180/PI);
end;


// URAL built-in functions
type
  TURALFuncConcat = class (TURALFunction)
  public
    function Execute (intr: TURALInterpreter; args: TURALStrList): ShortString; override;
  end;
  TURALFuncRound = class (TURALFunction)
  public
    function Execute (intr: TURALInterpreter; args: TURALStrList): ShortString; override;
  end;

  TURALFuncTrunc = class (TURALFunction)
  public
    function Execute (intr: TURALInterpreter; args: TURALStrList): ShortString; override;
  end;

  TURALFuncInt = class (TURALFunction)
  public
    function Execute (intr: TURALInterpreter; args: TURALStrList): ShortString; override;
  end;

  TURALFuncRnd = class (TURALFunction)
  public
    function Execute (intr: TURALInterpreter; args: TURALStrList): ShortString; override;
  end;

  TURALFuncAscii = class (TURALFunction)
  public
    function Execute (intr: TURALInterpreter; args: TURALStrList): ShortString; override;
  end;

  TURALFuncChr = class (TURALFunction)
  public
    function Execute (intr: TURALInterpreter; args: TURALStrList): ShortString; override;
  end;

  TURALFuncArgCount = class (TURALFunction)
  public
    function Execute (intr: TURALInterpreter; args: TURALStrList): ShortString; override;
  end;

  TURALFuncGetArg = class (TURALFunction)
  public
    function Execute (intr: TURALInterpreter; args: TURALStrList): ShortString; override;
  end;

  TURALFuncSinCosEtc = class (TURALFunction)
  protected
    fName: ShortString;
  public
    constructor Create (const pName: ShortString);

    function Execute (intr: TURALInterpreter; args: TURALStrList): ShortString; override;
  end;



{ TURALFuncRnd }
function TURALFuncRnd.Execute (intr: TURALInterpreter; args: TURALStrList): ShortString;
var
  error: Boolean;
  o0t: Integer;
  f0: Double;
begin
  if args.Count = 0 then
  begin
    result := URALFloat2Str(Random);
    exit;
  end;
  error := args.Count <> 1; if error then begin intr.Error('"RND": invalid argcount'); exit; end;
  o0t := URALTokenType(args[0]);
  if o0t = URTT_STRING then
  begin
    intr.Error('math with non-number');
  end
  else if o0t = URTT_FLOAT then
  begin
    f0 := URALStr2Float(args[0], error);
    try
      result := URALFloat2Str(Random*f0);
    except intr.Error('floating math error');
    end;
  end
  else result := URALInt2Str(Random(URALStr2Int(args[0], error)));
end;


{ TURALFuncRound }
function TURALFuncRound.Execute (intr: TURALInterpreter; args: TURALStrList): ShortString;
var
  error: Boolean;
  o0t: Integer;
  f0: Double;
begin
  error := args.Count <> 1; if error then begin intr.Error('"ROUND": invalid argcount'); exit; end;
  o0t := URALTokenType(args[0]);
  if o0t = URTT_STRING then
  begin
    intr.Error('math with non-number');
  end
  else if o0t = URTT_FLOAT then
  begin
    f0 := URALStr2Float(args[0], error);
    try
      result := URALInt2Str(round(f0));
    except intr.Error('floating math error');
    end;
  end
  else result := args[0];
end;


{ TURALFuncTrunc }
function TURALFuncTrunc.Execute (intr: TURALInterpreter; args: TURALStrList): ShortString;
var
  error: Boolean;
  o0t: Integer;
  f0: Double;
begin
  error := args.Count <> 1; if error then begin intr.Error('"TRUNC": invalid argcount'); exit; end;
  o0t := URALTokenType(args[0]);
  if o0t = URTT_STRING then
  begin
    intr.Error('math with non-number');
  end
  else if o0t = URTT_FLOAT then
  begin
    f0 := URALStr2Float(args[0], error);
    try
      result := URALInt2Str(trunc(f0));
    except intr.Error('floating math error');
    end;
  end
  else result := args[0];
end;


{ TURALFuncInt }
function TURALFuncInt.Execute (intr: TURALInterpreter; args: TURALStrList): ShortString;
var
  error: Boolean;
  o0t: Integer;
  f0: Double;
begin
  error := args.Count <> 1; if error then begin intr.Error('"INT": invalid argcount'); exit; end;
  o0t := URALTokenType(args[0]);
  if o0t = URTT_STRING then
  begin
    intr.Error('math with non-number');
  end
  else if o0t = URTT_FLOAT then
  begin
    f0 := URALStr2Float(args[0], error);
    try
      result := URALFloat2Str(trunc(f0));
    except intr.Error('floating math error');
    end;
  end
  else result := args[0]+'.0';
end;


{ TURALFuncConcat }
function TURALFuncConcat.Execute (intr: TURALInterpreter; args: TURALStrList): ShortString;
var
  f: Integer;
begin
  result[0] := #0;
  for f := 0 to args.Count-1 do result := result+args[f];
end;


{ TURALFuncAscii }
function TURALFuncAscii.Execute (intr: TURALInterpreter; args: TURALStrList): ShortString;
var
  error: Boolean;
begin
  error := args.Count <> 1; if error then begin intr.Error('"ASCII": invalid argcount'); exit; end;
  if Length(args[0]) = 0 then result := '-1'
  else result := URALInt2Str(ord(args[0][1]));
end;


{ TURALFuncChr }
function TURALFuncChr.Execute (intr: TURALInterpreter; args: TURALStrList): ShortString;
var
  error: Boolean;
begin
  error := args.Count <> 1; if error then begin intr.Error('"CHR": invalid argcount'); exit; end;
  if URALTokenType(args[0]) <> URTT_INT then intr.Error('"CHR": integer expected')
  else result := chr(URALStr2Int(args[0], error) and $FF);
end;


{ TURALFuncArgCount }
function TURALFuncArgCount.Execute (intr: TURALInterpreter; args: TURALStrList): ShortString;
begin
  if intr.fCurArgs = nil then result := '0'
  else result := URALInt2Str(intr.fCurArgs.Count);
end;


{ TURALFuncGetArg }
function TURALFuncGetArg.Execute (intr: TURALInterpreter; args: TURALStrList): ShortString;
var
  n: Integer;
  error: Boolean;
begin
  if args.Count <> 1 then begin intr.Error('"GETARG": invalid argcount'); exit; end;
  if URALTokenType(args[0]) <> URTT_INT then
  begin
    intr.Error('"GETARG": integer expected');
    exit;
  end;
  n := URALStr2Int(args[0], error);
  if n < 0 then result[0] := #0
  else if n = 0 then result := intr.fCurFuncName
  else if (intr.fCurArgs = nil) or (n > intr.fCurArgs.Count) then result[0] := #0
  else result := intr.fCurArgs[n-1];
end;


{ TURALFuncSinCosEtc }
constructor TURALFuncSinCosEtc.Create (const pName: ShortString);
begin
  inherited Create();
  fName := pName;
end;

function TURALFuncSinCosEtc.Execute (intr: TURALInterpreter; args: TURALStrList): ShortString;
var
  error: Boolean;
  o0t: Integer;
  f0: Double;
begin
  if args.Count <> 1 then
  begin
    intr.Error('"'+fName+'": invalid argcount'); exit;
  end;
  o0t := URALTokenType(args[0]);
  if o0t = URTT_STRING then
  begin
    intr.Error('"'+fName+'": number argument expected');
    exit;
  end;
  f0 := URALStr2Float(args[0], error);
  try
    if fName = 'SIN' then f0 := sin(f0)
    else if fName = 'COS' then f0 := cos(f0)
    else if fName = 'SQRT' then f0 := sqrt(f0)
    else if fName = 'TAN' then f0 := URALTan(f0)
    else if fName = 'EXP' then f0 := exp(f0)
    else if fName = 'LN' then f0 := ln(f0)
    else if fName = 'ARCSIN' then f0 := URALArcSin(f0)
    else if fName = 'ARCCOS' then f0 := URALArcCos(f0)
    else if fName = 'ARCTAN' then f0 := ArcTan(f0)
    else if fName = 'RAD2DEG' then f0 := URALRadToDeg(f0)
    else if fName = 'DEG2RAD' then f0 := URALDegToRad(f0)
    else
    begin
      intr.Error('unknown math: "'+fName+'"');
      exit;
    end;
    result := URALFloat2Str(f0);
  except intr.Error('floating math error');
  end;
end;


{ TURALInterpreter }
constructor TURALInterpreter.Create ();
begin
  inherited;
  fFuncList := TURALStrList.Create(true);
  fExtFuncs := TURALStrList.Create(true);
  SetLength(fVarPool, 0); fVarPoolUsed := 0;
  fError := false; fErrorMsg[0] := #0;
  fInstructionLimit := 10000000;
  RegisterStandardFunctions();
end;

destructor TURALInterpreter.Destroy ();
begin
  Clear(); ClearExternalFunc();
  SetLength(fVarPool, 0); fVarPoolUsed := 0;
  URALFreeAndNil(fExtFuncs);
  URALFreeAndNil(fFuncList);
  inherited;
end;

function TURALInterpreter.DoMathPlus (const a0, a1: ShortString): ShortString;
var
  err: Boolean;
  o0t, o1t: Integer;
  f0, f1: Double;
begin
  o0t := URALTokenType(a0); o1t := URALTokenType(a1);
  if (o0t = URTT_STRING) or (o1t = URTT_STRING) then
  begin
    result := a0+a1;
  end
  else if (o0t = URTT_FLOAT) or (o1t = URTT_FLOAT) then
  begin
    f0 := URALStr2Float(a0, err); f1 := URALStr2Float(a1, err);
    try
      result := URALFloat2Str(f0+f1);
    except Error('floating math err');
    end;
  end
  else
  begin
    o0t := URALStr2Int(a0, err); o1t := URALStr2Int(a1, err);
    result := URALInt2Str(o0t+o1t);
  end;
end;

function TURALInterpreter.DoMathMinus (const a0, a1: ShortString): ShortString;
var
  err: Boolean;
  o0t, o1t: Integer;
  f0, f1: Double;
begin
  o0t := URALTokenType(a0); o1t := URALTokenType(a1);
  if (o0t = URTT_STRING) or (o1t = URTT_STRING) then
  begin
    Error('math with non-number');
  end
  else if (o0t = URTT_FLOAT) or (o1t = URTT_FLOAT) then
  begin
    f0 := URALStr2Float(a0, err); f1 := URALStr2Float(a1, err);
    try
      result := URALFloat2Str(f0-f1);
    except Error('floating math err');
    end;
  end
  else
  begin
    o0t := URALStr2Int(a0, err); o1t := URALStr2Int(a1, err);
    result := URALInt2Str(o0t-o1t);
  end;
end;

function TURALInterpreter.DoMathMul (const a0, a1: ShortString): ShortString;
var
  err: Boolean;
  o0t, o1t: Integer;
  f0, f1: Double;
begin
  o0t := URALTokenType(a0); o1t := URALTokenType(a1);
  if (o0t = URTT_STRING) or (o1t = URTT_STRING) then
  begin
    Error('math with non-number');
  end
  else if (o0t = URTT_FLOAT) or (o1t = URTT_FLOAT) then
  begin
    f0 := URALStr2Float(a0, err); f1 := URALStr2Float(a1, err);
    try
      result := URALFloat2Str(f0*f1);
    except Error('floating math err');
    end;
  end
  else
  begin
    o0t := URALStr2Int(a0, err); o1t := URALStr2Int(a1, err);
    result := URALInt2Str(o0t*o1t);
  end;
end;

function TURALInterpreter.DoMathDiv (const a0, a1: ShortString): ShortString;
var
  err: Boolean;
  o0t, o1t: Integer;
  f0, f1: Double;
begin
  o0t := URALTokenType(a0); o1t := URALTokenType(a1);
  if (o0t = URTT_STRING) or (o1t = URTT_STRING) then
  begin
    Error('math with non-number');
  end
  else if (o0t = URTT_FLOAT) or (o1t = URTT_FLOAT) then
  begin
    f0 := URALStr2Float(a0, err); f1 := URALStr2Float(a1, err);
    try
      result := URALFloat2Str(f0/f1);
    except Error('floating math err');
    end;
  end
  else
  begin
    o0t := URALStr2Int(a0, err); o1t := URALStr2Int(a1, err);
    try
      result := URALFloat2Str(o0t/o1t);
    except Error('floating math err');
    end;
  end;
end;

function TURALInterpreter.DoMathIDiv (const a0, a1: ShortString): ShortString;
var
  err: Boolean;
  o0t, o1t: Integer;
  f0, f1: Double;
begin
  o0t := URALTokenType(a0); o1t := URALTokenType(a1);
  if (o0t = URTT_STRING) or (o1t = URTT_STRING) then
  begin
    Error('math with non-number');
  end
  else if (o0t = URTT_FLOAT) or (o1t = URTT_FLOAT) then
  begin
    f0 := URALStr2Float(a0, err); f1 := URALStr2Float(a1, err);
    try
      result := URALFloat2Str(int(f0)/int(f1));
    except Error('floating math err');
    end;
  end
  else
  begin
    o0t := URALStr2Int(a0, err); o1t := URALStr2Int(a1, err);
    try
      result := URALInt2Str(o0t div o1t);
    except Error('divizion by zero');
    end;
  end;
end;

function TURALInterpreter.DoMathMod (const a0, a1: ShortString): ShortString;
var
  err: Boolean;
  o0t, o1t: Integer;
  f0, f1: Double;
begin
  o0t := URALTokenType(a0); o1t := URALTokenType(a1);
  if (o0t = URTT_STRING) or (o1t = URTT_STRING) then
  begin
    Error('math with non-number');
  end
  else if (o0t = URTT_FLOAT) or (o1t = URTT_FLOAT) then
  begin
    f0 := URALStr2Float(a0, err); f1 := URALStr2Float(a1, err);
    try
      result := URALInt2Str(trunc(f0) mod trunc(f1));
    except err := true;
    end;
  end
  else
  begin
    o0t := URALStr2Int(a0, err); o1t := URALStr2Int(a1, err);
    try
      result := URALInt2Str(o0t mod o1t);
    except Error('divizion by zero');
    end;
  end;
end;

function TURALInterpreter.DoMathConcat (const a0, a1: ShortString): ShortString;
begin
  result := a0+a1;
end;

function TURALInterpreter.DoMathLess (const a0, a1: ShortString): ShortString;
var
  error: Boolean;
  o0t, o1t: Integer;
  f0, f1: Double;
begin
  o0t := URALTokenType(a0); o1t := URALTokenType(a1);
  if (o0t = URTT_STRING) or (o1t = URTT_STRING) then result := URALBool2Str(a0 < a1)
  else if (o0t = URTT_FLOAT) or (o1t = URTT_FLOAT) then
  begin
    f0 := URALStr2Float(a0, error); f1 := URALStr2Float(a1, error);
    result := URALBool2Str(f0 < f1);
  end
  else
  begin
    o0t := URALStr2Int(a0, error); o1t := URALStr2Int(a1, error);
    result := URALBool2Str(o0t < o1t);
  end;
end;

function TURALInterpreter.DoMathGreat (const a0, a1: ShortString): ShortString;
var
  error: Boolean;
  o0t, o1t: Integer;
  f0, f1: Double;
begin
  o0t := URALTokenType(a0); o1t := URALTokenType(a1);
  if (o0t = URTT_STRING) or (o1t = URTT_STRING) then result := URALBool2Str(a0 > a1)
  else if (o0t = URTT_FLOAT) or (o1t = URTT_FLOAT) then
  begin
    f0 := URALStr2Float(a0, error); f1 := URALStr2Float(a1, error);
    result := URALBool2Str(f0 > f1);
  end
  else
  begin
    o0t := URALStr2Int(a0, error); o1t := URALStr2Int(a1, error);
    result := URALBool2Str(o0t > o1t);
  end;
end;

function TURALInterpreter.DoMathEqu (const a0, a1: ShortString): ShortString;
var
  error: Boolean;
  o0t, o1t: Integer;
  f0, f1: Double;
begin
  o0t := URALTokenType(a0); o1t := URALTokenType(a1);
  if (o0t = URTT_STRING) or (o1t = URTT_STRING) then result := URALBool2Str(a0 = a1)
  else if (o0t = URTT_FLOAT) or (o1t = URTT_FLOAT) then
  begin
    f0 := URALStr2Float(a0, error); f1 := URALStr2Float(a1, error);
    result := URALBool2Str(f0 = f1);
  end
  else
  begin
    o0t := URALStr2Int(a0, error); o1t := URALStr2Int(a1, error);
    result := URALBool2Str(o0t = o1t);
  end;
end;

function TURALInterpreter.DoMathGreatEqu (const a0, a1: ShortString): ShortString;
var
  error: Boolean;
  o0t, o1t: Integer;
  f0, f1: Double;
begin
  o0t := URALTokenType(a0); o1t := URALTokenType(a1);
  if (o0t = URTT_STRING) or (o1t = URTT_STRING) then result := URALBool2Str(a0 >= a1)
  else if (o0t = URTT_FLOAT) or (o1t = URTT_FLOAT) then
  begin
    f0 := URALStr2Float(a0, error); f1 := URALStr2Float(a1, error);
    result := URALBool2Str(f0 >= f1);
  end
  else
  begin
    o0t := URALStr2Int(a0, error); o1t := URALStr2Int(a1, error);
    result := URALBool2Str(o0t >= o1t);
  end;
end;

function TURALInterpreter.DoMathLessEqu (const a0, a1: ShortString): ShortString;
var
  error: Boolean;
  o0t, o1t: Integer;
  f0, f1: Double;
begin
  o0t := URALTokenType(a0); o1t := URALTokenType(a1);
  if (o0t = URTT_STRING) or (o1t = URTT_STRING) then result := URALBool2Str(a0 <= a1)
  else if (o0t = URTT_FLOAT) or (o1t = URTT_FLOAT) then
  begin
    f0 := URALStr2Float(a0, error); f1 := URALStr2Float(a1, error);
    result := URALBool2Str(f0 <= f1);
  end
  else
  begin
    o0t := URALStr2Int(a0, error); o1t := URALStr2Int(a1, error);
    result := URALBool2Str(o0t <= o1t);
  end;
end;

function TURALInterpreter.DoMathNotEqu (const a0, a1: ShortString): ShortString;
var
  error: Boolean;
  o0t, o1t: Integer;
  f0, f1: Double;
begin
  o0t := URALTokenType(a0); o1t := URALTokenType(a1);
  if (o0t = URTT_STRING) or (o1t = URTT_STRING) then result := URALBool2Str(a0 <> a1)
  else if (o0t = URTT_FLOAT) or (o1t = URTT_FLOAT) then
  begin
    f0 := URALStr2Float(a0, error); f1 := URALStr2Float(a1, error);
    result := URALBool2Str(f0 <> f1);
  end
  else
  begin
    o0t := URALStr2Int(a0, error); o1t := URALStr2Int(a1, error);
    result := URALBool2Str(o0t <> o1t);
  end;
end;

function TURALInterpreter.DoMathLogAnd (const a0, a1: ShortString): ShortString;
begin
  result := URALBool2Str(URALIsTrue(a0) and URALIsTrue(a1));
end;

function TURALInterpreter.DoMathLogOr (const a0, a1: ShortString): ShortString;
begin
  result := URALBool2Str(URALIsTrue(a0) or URALIsTrue(a1));
end;

function TURALInterpreter.DoMathLogXor (const a0, a1: ShortString): ShortString;
begin
  result := URALBool2Str(URALIsTrue(a0) xor URALIsTrue(a1));
end;

function TURALInterpreter.DoMathLogNot (const a0, a1: ShortString): ShortString;
begin
  result := URALBool2Str(not URALIsTrue(a0));
end;

function TURALInterpreter.DoMathBitNot (const a0, a1: ShortString): ShortString;
var
  err: Boolean;
  o0t: Integer;
begin
  o0t := URALTokenType(a0);
  if o0t <> URTT_INT then Error('math with non-number')
  else
  begin
    o0t := URALStr2Int(a0, err);
    result := URALInt2Str(not o0t);
  end;
end;

function TURALInterpreter.DoMathBitAnd (const a0, a1: ShortString): ShortString;
var
  err: Boolean;
  o0t, o1t: Integer;
begin
  o0t := URALTokenType(a0); o1t := URALTokenType(a0);
  if (o0t <> URTT_INT) or (o1t <> URTT_INT) then Error('math with non-number')
  else
  begin
    o0t := URALStr2Int(a0, err); o1t := URALStr2Int(a1, err);
    result := URALInt2Str(o0t and o1t);
  end;
end;

function TURALInterpreter.DoMathBitOr (const a0, a1: ShortString): ShortString;
var
  err: Boolean;
  o0t, o1t: Integer;
begin
  o0t := URALTokenType(a0); o1t := URALTokenType(a0);
  if (o0t <> URTT_INT) or (o1t <> URTT_INT) then Error('math with non-number')
  else
  begin
    o0t := URALStr2Int(a0, err); o1t := URALStr2Int(a1, err);
    result := URALInt2Str(o0t or o1t);
  end;
end;

function TURALInterpreter.DoMathBitXor (const a0, a1: ShortString): ShortString;
var
  err: Boolean;
  o0t, o1t: Integer;
begin
  o0t := URALTokenType(a0); o1t := URALTokenType(a0);
  if (o0t <> URTT_INT) or (o1t <> URTT_INT) then Error('math with non-number')
  else
  begin
    o0t := URALStr2Int(a0, err); o1t := URALStr2Int(a1, err);
    result := URALInt2Str(o0t xor o1t);
  end;
end;



procedure TURALInterpreter.RegisterStandardFunctions ();
begin
  fStdMath[KW_CONCAT] := DoMathConcat;
  fStdMath[KW_PLUS] := DoMathPlus;
  fStdMath[KW_MINUS] := DoMathMinus;
  fStdMath[KW_MUL] := DoMathMul;
  fStdMath[KW_DIV] := DoMathDiv;
  fStdMath[KW_IDIV] := DoMathIDiv;

  fStdMath[KW_LOGNOT] := DoMathLogNot;
  fStdMath[KW_BITNOT] := DoMathBitNot;

  fStdMath[KW_BITAND] := DoMathBitAnd;
  fStdMath[KW_BITOR] := DoMathBitOr;
  fStdMath[KW_BITXOR] := DoMathBitXor;

  fStdMath[KW_LESS] := DoMathLess;
  fStdMath[KW_LESSEQU] := DoMathLessEqu;
  fStdMath[KW_GREAT] := DoMathGreat;
  fStdMath[KW_GREATEQU] := DoMathGreatEqu;
  fStdMath[KW_EQU] := DoMathEqu;
  fStdMath[KW_NOTEQU] := DoMathNotEqu;

  fStdMath[KW_LOGAND] := DoMathLogAnd;
  fStdMath[KW_LOGOR] := DoMathLogOr;
  fStdMath[KW_LOGXOR] := DoMathLogXor;

  fStdMath[KW_MOD] := DoMathMod;

  AddExternalFunc('ROUND', TURALFuncRound.Create());
  AddExternalFunc('TRUNC', TURALFuncTrunc.Create());
  AddExternalFunc('INT', TURALFuncInt.Create());
  AddExternalFunc('CONCAT', TURALFuncConcat.Create());
  AddExternalFunc('RND', TURALFuncRnd.Create());
  AddExternalFunc('ASCII', TURALFuncAscii.Create());
  AddExternalFunc('CHR', TURALFuncChr.Create());
  AddExternalFunc('ARGCOUNT', TURALFuncArgCount.Create());
  AddExternalFunc('GETARG', TURALFuncGetArg.Create());

  AddExternalFunc('SIN', TURALFuncSinCosEtc.Create('SIN'));
  AddExternalFunc('COS', TURALFuncSinCosEtc.Create('COS'));
  AddExternalFunc('TAN', TURALFuncSinCosEtc.Create('TAN'));
  AddExternalFunc('SQRT', TURALFuncSinCosEtc.Create('SQRT'));
  AddExternalFunc('EXP', TURALFuncSinCosEtc.Create('EXP'));
  AddExternalFunc('LN', TURALFuncSinCosEtc.Create('LN'));
  AddExternalFunc('ARCSIN', TURALFuncSinCosEtc.Create('ARCSIN'));
  AddExternalFunc('ARCCOS', TURALFuncSinCosEtc.Create('ARCCOS'));
  AddExternalFunc('ARCTAN', TURALFuncSinCosEtc.Create('ARCTAN'));
  AddExternalFunc('RAD2DEG', TURALFuncSinCosEtc.Create('RAD2DEG'));
  AddExternalFunc('DEG2RAD', TURALFuncSinCosEtc.Create('DEG2RAD'));
end;

procedure TURALInterpreter.Error (const msg: ShortString);
var
  f, n, e: Integer;
begin
  if fError then exit;
  fError := true; fErrorMsg := msg;
  fErrorFunc := fCurFuncName;
  // find error position
  fErrorLine := 1;
  if fCurFunc = nil then exit;
  f := fCurIP;
  while f >= 0 do
  begin
    if (fCurFunc.Objects[f] = TK_EOP) and
       (fCurFunc[f] <> '') then
    begin
      Val(fCurFunc[f], n, e);
      if e = 0 then begin fErrorLine := n; exit; end;
    end;
    Dec(f);
  end;
end;

procedure TURALInterpreter.Clear ();
var
  f: Integer;
  sl: TURALStrList;
begin
  for f := 0 to fFuncList.Count-1 do
  begin
    sl := TURALStrList(fFuncList.Objects[f]);
    sl.Objects[0].Free(); // if any...
    fFuncList.Objects[f] := nil; sl.Free();
  end;
  fFuncList.Clear();
end;

procedure TURALInterpreter.ClearExternalFunc ();
var
  f: Integer;
begin
  for f := 0 to fExtFuncs.Count-1 do
  begin
    fExtFuncs.Objects[f].Free(); fExtFuncs.Objects[f] := nil;
  end;
  fExtFuncs.Clear();
end;

function TURALInterpreter.PrecompileFunction (pFunc: TURALStrList;
  out cfn: TURALStrList; tokenized: Boolean = false): ShortString;
var
  f: Integer;
  {$IFDEF XCOMPILE}
  e: Integer;
  ss: ShortString;
  {$ENDIF}
  args: TURALStrList;
begin
  result[0] := #0; cfn := nil;
  if pFunc = nil then exit;
  if tokenized then
  begin
    cfn := TURALStrList.Create();
    try
      for f := 0 to pFunc.Count-1 do cfn.AddObject(pFunc[f], pFunc.Objects[f]);
    except
      cfn.Free();
      raise;
    end;
    pFunc := cfn; cfn := nil;
  end
  else pFunc := Tokenize(pFunc);
  try
    args := TURALStrList.Create();
  except
    pFunc.Free();
    raise;
  end;
  try
    // prepare header
    f := 0;
    // skip leading empty lines
    while (f < pFunc.Count) and (pFunc.Objects[f] = TK_EOP) do Inc(f);
    // "SUB"?
    if (pFunc.Objects[f] = TK_KEYWORD) and (pFunc[f] = KW_SUB) then
    begin
      Inc(f);
      // here must be name or "("
      if (pFunc.Objects[f] = TK_ID) or
         ((pFunc.Objects[f] = TK_KEYWORD) and (pFunc[f][1] = '(')) then
      begin
        if pFunc.Objects[f] = TK_ID then
        begin
          result := pFunc[f]; Inc(f);
        end;
        // here must be "(" or EOL
        if (pFunc.Objects[f] = TK_KEYWORD) and (pFunc[f] = '(') then
        begin
          // collect arguments
          Inc(f);
          if (pFunc.Objects[f] <> TK_KEYWORD) or (pFunc[f] <> ')') then
          begin
            repeat
              if pFunc.Objects[f] <> TK_ID then begin f := pFunc.Count; break; end;
              args.Add(pFunc[f]); Inc(f);
              if pFunc.Objects[f] <> TK_KEYWORD then begin f := pFunc.Count; break; end;
              if pFunc[f] = ')' then begin Inc(f); break; end;
              if pFunc[f] <> ',' then begin f := pFunc.Count; break; end;
              Inc(f);
            until false;
          end
          else
          begin
            if (pFunc.Objects[f] = TK_KEYWORD) and (pFunc[f] = ')') then
            begin
              Inc(f);
              // here must be EOL
              if pFunc.Objects[f] <> TK_EOP then
              begin
                pFunc.Insert(f, '0'); pFunc.Objects[f] := TK_EOP;
              end;
            end
            else f := pFunc.Count;
          end;
        end
        else
        begin
          // here must be EOL
          if pFunc.Objects[f] <> TK_EOP then f := pFunc.Count;
        end;
        if f < pFunc.Count then
        begin
          // delete header
          while f > 0 do begin pFunc.Delete(0); Dec(f); end;
          // insert arg names
          for f := args.Count-1 downto 0 do
          begin
            pFunc.Insert(0, args[f]); pFunc.Objects[0] := TK_ID;
          end;
          // insert execution counter & deletion flag
          pFunc.Insert(0, #0#0#0#0);
          PInteger(@(pFunc.fItems^[0]))^ := 0;
          f := 0;
        end;
      end
      else f := pFunc.Count;
    end
    else f := pFunc.Count;
    args.Free();
    if f >= pFunc.Count then
    begin
      // this is not a function
      pFunc.Free();
      exit;
    end;
    f := 1; while (f < pFunc.Count) and (pFunc.Objects[f] <> TK_EOP) do Inc(f);
    {$IFDEF XCOMPILE}
    while f < pFunc.Count do
    begin
      if pFunc.Objects[f] = TK_ID then
      begin
        e := fExtFuncs.IndexOf(pFunc[f]);
        if e <> -1 then
        begin
          Move(e, ss[1], 4); ss[0] := #4;
          pFunc[f] := #0+ss+pFunc[f];
        end;
      end;
      Inc(f);
    end;
    {$ENDIF}
    cfn := pFunc;
  except
    pFunc.Free(); cfn := nil; result[0] := #0;
    raise;
  end;
end;

function TURALInterpreter.ShortStringToStrList (const st: ShortString): TURALStrList;
var
  f, e: Integer;
begin
  result := TURALStrList.Create();
  try
    f := 1; while f <= Length(st) do
    begin
      e := f; while (e <= Length(st)) and not (st[e] in [#10,#13]) do Inc(e);
      result.Add(Copy(st, f, e-f)); Inc(e);
      if (e <= Length(st)) and (st[e-1] = #13) and (st[e] = #10) then Inc(e);
      f := e;
    end;
  except
    result.Free();
    raise;
  end;
end;

function TURALInterpreter.AnsiStringToStrList (const st: AnsiString): TURALStrList;
var
  f, e: Integer;
begin
  result := TURALStrList.Create();
  try
    f := 1; while f <= Length(st) do
    begin
      e := f; while (e <= Length(st)) and not (st[e] in [#10,#13]) do Inc(e);
      if e-f > 255 then
      begin
        e := f+254; // FIXME!
        result.Add(Copy(st, f, e-f)); Inc(e);
      end
      else
      begin
        result.Add(Copy(st, f, e-f)); Inc(e);
        if (e <= Length(st)) and (st[e-1] = #13) and (st[e] = #10) then Inc(e);
      end;
      f := e;
    end;
  except
    result.Free();
    raise;
  end;
end;

procedure TURALInterpreter.AddFunction (const pFunc: ShortString);
var
  sl: TURALStrList;
begin
  sl := ShortStringToStrList(pFunc);
  try
    AddFunction(sl);
  finally
    sl.Free();
  end;
end;

procedure TURALInterpreter.AddFunctionLong (const pFunc: AnsiString);
var
  sl: TURALStrList;
begin
  sl := AnsiStringToStrList(pFunc);
  try
    AddFunction(sl);
  finally
    sl.Free();
  end;
end;

procedure TURALInterpreter.AddFunction (pFunc: TURALStrList;
  tokenized: Boolean = false);
var
  pName: ShortString;
  cfn: TURALStrList;
  t: TObject;
  f: Integer;
begin
  if pFunc = nil then exit;
  pName := PrecompileFunction(pFunc, cfn, tokenized);
  if (cfn = nil) or (pName[0] = #0) then exit;

  try
    f := fFuncList.IndexOf(pName);
    if f <> -1 then
    begin
      pFunc := TURALStrList(fFuncList.Objects[f]);
      if fInDoExecute and (PInteger(@(pFunc.fItems^[0]))^ <> 0) then
      begin
        // cache update
        t := pFunc.Objects[0]; pFunc.Objects[0] := cfn; t.Free();
        pFunc.fItems^[0][4] := #0; // clear deletion flag
      end
      else
      begin
        fFuncList.Objects[f].Free(); fFuncList.Objects[f] := cfn;
      end;
    end
    else fFuncList.AddObject(pName, cfn);
  except
    cfn.Free();
  end;
end;

function TURALInterpreter.HasFunction (const pName: ShortString): Boolean;
begin
  result := fFuncList.IndexOf(pName) <> -1;
end;

procedure TURALInterpreter.DeleteFunction (const pName: ShortString);
var
  f: Integer;
  pFunc: TURALStrList;
  t: TObject;
  fn: ShortString;
begin
  fn := pName; URALUpStr(fn);
  f := fFuncList.IndexOf(fn);
  if f <> -1 then
  begin
    pFunc := TURALStrList(fFuncList.Objects[f]);
    if fInDoExecute and (PInteger(@(pFunc.fItems^[0]))^ <> 0) then
    begin
      // cache update
      t := pFunc.Objects[0]; pFunc.Objects[0] := nil; t.Free();
      pFunc.fItems^[0][4] := #1; // set deletion flag
    end
    else
    begin
      fFuncList.Objects[f].Free(); fFuncList.Objects[f] := nil;
      fFuncList.Delete(f);
    end;
  end;
end;

procedure TURALInterpreter.AddExternalFunc (const pName: ShortString; pUFunc: TURALFunction);
var
  f: Integer;
  ss: ShortString;
begin
  if pUFunc = nil then exit;
  ss := pName; URALUpStr(ss);
  f := fExtFuncs.IndexOf(ss);
  if f <> -1 then
  begin
    fExtFuncs.Objects[f].Free(); fExtFuncs.Objects[f] := pUFunc;
  end
  else fExtFuncs.AddObject(ss, pUFunc);
end;

procedure TURALInterpreter.DeleteExternalFunc (const pName: ShortString);
var
  f: Integer;
  ss: ShortString;
begin
  ss := pName; URALUpStr(ss);
  f := fExtFuncs.IndexOf(ss);
  if f <> -1 then
  begin
    fExtFuncs.Objects[f].Free(); fExtFuncs.Objects[f] := nil;
    fExtFuncs.Delete(f);
  end;
end;

function TURALInterpreter.FindExternalFunc (const pName: ShortString): TURALFunction;
var
  ss: ShortString;
begin
  ss := pName; URALUpStr(ss);
  result := TURALFunction(fExtFuncs.Objects[fExtFuncs.IndexOf(ss)]);
  if Integer(result) < 1024 then result := nil;
end;

function TURALInterpreter.Tokenize (pFunc: TURALStrList): TURALStrList;
label specagain, mainloopstart;
type
  TCharSet = set of Char;
var
  line, pos, st: Integer;
  s, t: ShortString;
  qch, ch: Char;
  f, cc: Integer;
  tokenList: TURALStrList;

  function TokenCollect (const cs: TCharSet): ShortString;
  var
    st: Integer;
  begin
    st := pos;
    while (pos <= Length(s)) and (s[pos] in cs) do Inc(pos);
    result := Copy(s, st, pos-st);
  end;

  procedure AddToken (const tk: ShortString; ttype: Pointer; newLn: Boolean = false);
  begin
    tokenList.AddObject(tk, ttype);
  end;

begin
  if pFunc = nil then begin result := nil; exit; end;
  result := TURALStrList.Create(); tokenList := result;
  try
    // tokenize line by line
    line := 0; if line < pFunc.Count then s := pFunc[line] else s[0] := #0;
    pos := 1;
mainloopstart:
    while line < pFunc.Count do
    begin
      // skip spaces and comments
      while pos <= Length(s) do
      begin
        ch := s[pos];
        if ch = '/' then
        begin
          if (pos < Length(s)) and (s[pos+1] = '/') then pos := 256
          else if (pos < Length(s)) and (s[pos+1] = '*') then
          begin
            // skip multiline comment
            Inc(pos, 2);
            repeat
              while pos < Length(s) do
              begin
                if (s[pos] = '*') and (s[pos+1] = '/') then
                begin
                  Inc(pos, 2);
                  goto mainloopstart;
                end;
                Inc(pos);
              end;
              Inc(line);
              if line >= pFunc.Count then goto mainloopstart;
              AddToken(URALInt2Str(line), TK_EOP);
              s := pFunc[line]; pos := 1;
            until false;
            goto mainloopstart;
          end
          else break;
        end
        else if ch = '\' then
        begin
          if (pos < Length(s)) and (s[pos+1] = '\') then
          begin
            // go to the new line (line concatenation)
            pos := 1; Inc(line); s := pFunc[line];
            goto mainloopstart;
          end
          else break;
        end
        else if ch = #13 then
        begin
          if (pos < Length(s)) and (s[pos+1] = #10) then Inc(pos);
          s[pos] := ';'; break;
        end
        else if ch = #10 then
        begin
          s[pos] := ';'; break;
        end
        else if ch > ' ' then break;
        Inc(pos);
      end;
      //tkLine := line;
      if (pos > Length(s)) or (s[pos] = ';') then
      begin
        // new operator
        // add TK_EOP
        if pos > Length(s) then AddToken(URALInt2Str(line+1), TK_EOP)
        else AddToken('', TK_EOP);
        Inc(pos); // skip ";"
        // new line?
        if pos > Length(s) then
        begin
          Inc(line);
          if line >= pFunc.Count then break;
          s := pFunc[line]; pos := 1;
        end;
        continue;
      end;
      // collect token
      case s[pos] of
        '<':
          begin
            Inc(pos);
            if (pos <= Length(s)) and (s[pos] = '>') then
              begin Inc(pos); AddToken(KW_NOTEQU, TK_KEYWORD); end
            else if (pos <= Length(s)) and (s[pos] = '=') then
              begin Inc(pos); AddToken(KW_LESSEQU, TK_KEYWORD); end
            else AddToken(KW_LESS, TK_KEYWORD);
          end;
        '>':
          begin
            Inc(pos);
            if (pos <= Length(s)) and (s[pos] = '=') then
              begin Inc(pos); AddToken(KW_GREATEQU, TK_KEYWORD); end
            else AddToken(KW_GREAT, TK_KEYWORD);
          end;
        '|':
          begin
            Inc(pos);
            if (pos <= Length(s)) and (s[pos] = '|') then
              begin Inc(pos); AddToken(KW_CONCAT, TK_KEYWORD); end
            else AddToken(KW_LOGOR, TK_KEYWORD);
          end;
        '_','A'..'Z','a'..'z':
          begin
            st := pos;
            while (pos <= Length(s)) and
                  (s[pos] in ['0'..'9','_','A'..'Z','a'..'z']) do Inc(pos);
            t := Copy(s, st, pos-st);
            for f := 1 to Length(t) do t[f] := UpCase(t[f]);
            if t = 'SUB' then AddToken(KW_SUB, TK_KEYWORD)
            else if t = 'ENDSUB' then AddToken(KW_SUBEND, TK_KEYWORD)
            else if t = 'IF' then AddToken(KW_IF, TK_KEYWORD)
            else if t = 'THEN' then AddToken(KW_THEN, TK_KEYWORD)
            else if t = 'ELSE' then AddToken(KW_ELSE, TK_KEYWORD)
            else if t = 'ENDIF' then AddToken(KW_ENDIF, TK_KEYWORD)
            else if t = 'ELSIF' then AddToken(KW_ELSIF, TK_KEYWORD)
            else if t = 'GOELSE' then AddToken(KW_GOELSE, TK_KEYWORD)
            else if t = 'BREAKIF' then AddToken(KW_BREAKIF, TK_KEYWORD)
            else if t = 'WHILE' then AddToken(KW_WHILE, TK_KEYWORD)
            else if t = 'WEND' then AddToken(KW_WEND, TK_KEYWORD)
            else if t = 'AGAIN' then AddToken(KW_AGAIN, TK_KEYWORD)
            else if t = 'BREAK' then AddToken(KW_BREAK, TK_KEYWORD)
            else if t = 'RETURN' then AddToken(KW_RETURN, TK_KEYWORD)
            else if t = 'MOD' then AddToken(KW_MOD, TK_KEYWORD)
            else if t = 'AND' then AddToken(KW_BITAND, TK_KEYWORD)
            else if t = 'OR' then AddToken(KW_BITOR, TK_KEYWORD)
            else if t = 'XOR' then AddToken(KW_BITXOR, TK_KEYWORD)
            else AddToken(t, TK_ID);
          end;
      '"', #39:
        begin
          // string
          qch := s[pos];
          Inc(pos); t[0] := #0;
          while (pos <= Length(s)) do
          begin
            if s[pos] = qch then
            begin
              if pos = Length(s) then break;
              if s[pos+1] = qch then
              begin
                // add char
                Inc(t[0]); t[Byte(t[0])] := s[pos];
                Inc(pos, 2);
              end;
              if (s[pos+1] <> '$') and (s[pos+1] <> '^') then break;
              // special char
              Inc(pos, 2); // skip quote and type id
              if pos > Length(s) then break;
specagain:
              if s[pos-1] = '^' then
              begin
                // control char
                ch := UpCase(s[pos]);
                if (ch < 'A') or (ch > 'Z') then ch := 'A';
                t := t+Char(Byte(ch)-ord('A')+1);
                Inc(pos);
              end
              else if s[pos] in ['0'..'9'] then
              begin
                // char code
                cc := 0;
                while s[pos] in ['0'..'9'] do
                begin
                  cc := cc*10+ord(s[pos])-ord('0'); Inc(pos);
                end;
                t := t+Char(cc and $FF);
              end;
              if pos > Length(s) then break;
              if s[pos] in ['$','^'] then begin Inc(pos); goto specagain; end;
              if (s[pos] <> '"') and (s[pos] <> #39) then begin Dec(pos); break; end;
              qch := s[pos];
            end
            else
            begin
              // add char
              Inc(t[0]); t[Byte(t[0])] := s[pos];
              Inc(pos);
            end;
          end;
          Inc(pos);
          AddToken(t, TK_OTHER);
        end;
      '0'..'9','#','%','@','$':
        begin
          t[0] := #0;
          if s[pos] in ['0'..'9'] then
          begin
            t := TokenCollect(['0'..'9']);
            if (pos <= Length(s)) and (s[pos] = '.') then
            begin
              t := t+'.'; Inc(pos);
              if s[pos] in ['0'..'9'] then
              begin
                t := t+TokenCollect(['0'..'9']);
              end;
            end;
          end
          else
          begin
            t := s[pos]; Inc(pos);
            case t[1] of
              '%':
                if (pos <= Length(s)) and (s[pos] in ['0'..'1']) then
                  t := t+TokenCollect(['0'..'1']);
              '@':
                if (pos <= Length(s)) and (s[pos] in ['0'..'7']) then
                  t := t+TokenCollect(['0'..'7']);
              '#','$':
                if (pos <= Length(s)) and (s[pos] in ['0'..'9','A'..'F','a'..'f']) then
                begin
                  t := t+TokenCollect(['0'..'9','A'..'F','a'..'f']);
                  for f := 2 to Length(t) do t[f] := UpCase(t[f]);
                end;
            end;
          end;
          AddToken(t, TK_OTHER);
        end;
        '!': begin AddToken(KW_LOGNOT, TK_KEYWORD); Inc(pos); end;
        '&': begin AddToken(KW_LOGAND, TK_KEYWORD); Inc(pos); end;
        '^': begin AddToken(KW_LOGXOR, TK_KEYWORD); Inc(pos); end;
        '+': begin AddToken(KW_PLUS, TK_KEYWORD); Inc(pos); end;
        '-': begin AddToken(KW_MINUS, TK_KEYWORD); Inc(pos); end;
        '*': begin AddToken(KW_MUL, TK_KEYWORD); Inc(pos); end;
        '/': begin AddToken(KW_DIV, TK_KEYWORD); Inc(pos); end;
        '\': begin AddToken(KW_IDIV, TK_KEYWORD); Inc(pos); end;
        '=': begin AddToken(KW_EQU, TK_KEYWORD); Inc(pos); end;
        '~': begin AddToken(KW_BITNOT, TK_KEYWORD); Inc(pos); end;
        '[': begin AddToken(s[pos], TK_KEYWORD); Inc(pos); end;
        ']': begin AddToken(s[pos], TK_KEYWORD); Inc(pos); end;
        '(': begin AddToken(s[pos], TK_KEYWORD); Inc(pos); end;
        ')': begin AddToken(s[pos], TK_KEYWORD); Inc(pos); end;
        '.': begin AddToken(s[pos], TK_KEYWORD); Inc(pos); end;
        ':': begin AddToken(s[pos], TK_KEYWORD); Inc(pos); end;
        ',': begin AddToken(s[pos], TK_KEYWORD); Inc(pos); end;
        #127..#255: begin AddToken(s[pos], TK_OTHER); Inc(pos); end;
        else AddToken(s[pos], TK_OTHER); Inc(pos);
      end;
    end;
    // drop out all leading empty lines and operator delimiters
    while (result.Count > 0) and (result.Objects[0] = TK_EOP) do result.Delete(0);
    // drop out all unnecessary operator delimiters
    f := 0; while f < result.Count do
    begin
      if (result.Objects[f] = TK_EOP) and (result[f] = '') then
      begin
        // operator delimiter
        // check if next or previous token is eol or eop
        if (result.Objects[f+1] = TK_EOP) or
           (result.Objects[f-1] = TK_EOP) then
        begin
          // we can delete this delimiter
          result.Delete(f);
        end
        else Inc(f);
      end
      else Inc(f);
    end;
    // drop out all unnecessary eols
    f := 0; while f < result.Count-1 do
    begin
      if (result.Objects[f] = TK_EOP) and
         (result[f] <> '') then
      begin
        // eol
        // check if next token is eol
        if (result.Objects[f+1] = TK_EOP) and
           (result[f+1] <> '') then
        begin
          // we can delete this eol
          result.Delete(f);
        end
        else Inc(f);
      end
      else Inc(f);
    end;
    //result.AddObject('', TK_EOP);
    //result.AddObject(KW_SUBEND, TK_KEYWORD);
  except
    result.Free();
  end;
end;

function Uncode (const s: string): string;
begin
  result := s;
  if (Length(result) > 4) and (result[1] = #0) then Delete(result, 1, 5);
end;

function TURALInterpreter.CollectTrain (const leaderName: ShortString): ShortString;
// "." already skipped!
var
  tt: Pointer;
  r0: ShortString;
begin
  result := leaderName; Dec(fCurIP); r0[0] := #0;
  repeat
    Inc(fCurIP); // skip "."
    result := result+'.';
    tt := fCurFunc.Objects[fCurIP];
    if tt = TK_ID then
    begin
      {$IFDEF XCOMPILE}
      r0 := fCurFunc[fCurIP];
      if (Length(r0) > 4) and (r0[1] = #0) then Delete(r0, 1, 5);
      r0 := GetLocalVar(r0); Inc(fCurIP);
      {$ELSE}
      r0 := GetLocalVar(fCurFunc[fCurIP]); Inc(fCurIP);
      {$ENDIF}
    end
    else if tt = TK_OTHER then
    begin
      r0 := fCurFunc[fCurIP]; Inc(fCurIP);
    end
    else if (tt = TK_KEYWORD) and (fCurFunc[fCurIP][1] = '(') then
    begin
      r0 := Expression(); if fError then begin result[0] := #0; exit; end;
    end
    else
    begin
      result[0] := #0;
      Error('".": identifier expected');
      exit;
    end;
    URALUpStr(r0);
    result := result+r0;
  until (fCurFunc.Objects[fCurIP] <> TK_KEYWORD) or
        (fCurFunc[fCurIP][1] <> '.');
end;

procedure TURALInterpreter.ParseFunctionArgs (al: TURALStrList);
// "(" already skipped!
var
  r0: ShortString;
begin
  if (fCurFunc.Objects[fCurIP] = TK_KEYWORD) and (fCurFunc[fCurIP][1] = ')') then
  begin
    // no args, just skip ")"
    Inc(fCurIP);
  end
  else
  begin
    // collect args
    repeat
      r0 := Expression(); if fError then exit;
      al.Add(r0);
      if fCurFunc.Objects[fCurIP] = TK_KEYWORD then
      begin
        Inc(fCurIP); // skip token
        case fCurFunc[fCurIP-1][1] of
          ')': break; // no more args
          ',': continue;
        end;
        // if we got here then something is wrong
        Dec(fCurIP); // get back to the bad token
      end;
      Error('"," or ")" expected');
      exit;
    until false;
  end;
end;

function TURALInterpreter.DoFunctionCall (const funcname: ShortString): ShortString;
// "(" already skipped
var
  al: TURALStrList;
  uf: TURALFunction;
  {$IFDEF XCOMPILE}
  f: Integer;
  {$ENDIF}
begin
  result[0] := #0;
  // collect args
  al := TURALStrList.Create(false);
  try
    ParseFunctionArgs(al); if fError then exit;
    {$IFDEF XCOMPILE}
    if (Length(funcname) > 4) and (funcname[1] = #0) then
    begin
      Move(funcname[2], f, 4);
      uf := TURALFunction(fExtFuncs.Objects[f]);
    end
    else uf := FindExternalFunc(funcname);
    {$ELSE}
    uf := FindExternalFunc(funcname);
    {$ENDIF}
    if uf <> nil then result := uf.ExecuteEx(self, al, funcname)
    else result := DoExecute(funcname, al);
  finally
    al.Free();
  end;
end;

function TURALInterpreter.Expression (): ShortString;

  procedure DoStringIndexing (var res: ShortString);
  // "[" already skipped!
  var
    a, b: Integer;
    err: Boolean;
    r0: ShortString;
  begin
    if (fCurFunc.Objects[fCurIP] = TK_KEYWORD) and (fCurFunc[fCurIP][1] = ':') then
    begin
      // no a
      a := 1;
    end
    else
    begin
      r0 := Expression(); if fError then begin res[0] := #0; exit; end;
      if URALTokenType(r0) <> URTT_INT then
      begin
        res[0] := #0;
        Error('"[]": integer expected');
        exit;
      end;
      a := URALStr2Int(r0, err);
    end;
    if (fCurFunc.Objects[fCurIP] = TK_KEYWORD) and (fCurFunc[fCurIP][1] = ']') then
    begin
      // no b
      if a = 0 then
      begin
        // we want length
        res := URALInt2Str(Length(res));
      end
      else
      begin
        // we want char
        if (a < 1) or (a > Length(res)) then res[0] := #0 else res := res[a];
      end;
      Inc(fCurIP);
      exit;
    end;
    if (fCurFunc.Objects[fCurIP] <> TK_KEYWORD) or (fCurFunc[fCurIP][1] <> ':') then
    begin
      res[0] := #0;
      Error('"[]": ":" expected');
      exit;
    end;
    Inc(fCurIP); // skip ":"
    r0 := Expression(); if fError then begin res[0] := #0; exit; end;
    if URALTokenType(r0) <> URTT_INT then
    begin
      res[0] := #0;
      Error('"[]": integer expected');
      exit;
    end;
    b := URALStr2Int(r0, err);
    if b = 0 then res[0] := #0
    else
    begin
      if a < 1 then a := 1;
      if b < 0 then b := Length(res)+b+1;
      if b < a then res[0] := #0 else res := Copy(res, a, b-a+1);
    end;
    // skip "]"
    if (fCurFunc.Objects[fCurIP] <> TK_KEYWORD) or (fCurFunc[fCurIP][1] <> ']') then
    begin
      res[0] := #0;
      Error('"[]": "]" expected');
      exit;
    end;
    Inc(fCurIP);
  end;

  function DoEvalExpr (al: TURALStrList): ShortString;
  var
    oldIP: Integer;
    oldFn: TURALStrList;
  begin
    al := Tokenize(al); oldIP := fCurIP; oldFn := fCurFunc;
    try
      fCurFunc := al; fCurIP := 0;
      result := Expression();
    finally
      fCurFunc := oldFn; fCurIP := oldIP;
      if fError then
      begin
        // regenerate error
        fError := false;
        Error('invalid expression in EVALEXPR()');
      end;
      al.Free();
    end;
  end;

  function DoEvalSub (al: TURALStrList): ShortString;
  var
    fn: TURALStrList;
    errPos: Integer;
  begin
    if (fCurFunc.Objects[fCurIP] <> TK_KEYWORD) or
       (fCurFunc[fCurIP][1] <> '(') then
    begin
      Error('EXECSUB() can be only execed %-)');
      result[0] := #0;
      exit;
    end;
    PrecompileFunction(al, fn); al := nil; errPos := fCurIP;
    if fn = nil then
    begin
      Error('EXECSUB() can''t prepare sub');
      result[0] := #0;
      exit;
    end;
    try
      al := TURALStrList.Create(false);
      Inc(fCurIP); ParseFunctionArgs(al);
      if fError then
      begin
        al.Free(); al := nil;
      end
      else result := DoExecute('', al, fn);
    finally
      if fError and (al = nil) then
      begin
        // regenerate error
        fError := false; fCurIP := errPos;
        Error('invalid sub in EXECSUB()');
      end;
      al.Free(); fn.Free();
    end;
  end;

  function DoExpr (prio: Integer): ShortString;
  var
    r0, t, funcname: ShortString;
    ch, opch: Char;
    f: Integer;
    al: TURALStrList;
    uf: TURALFunction;
    tt: Pointer;
  begin
    if prio = 0 then
    begin
      // term
      tt := fCurFunc.Objects[fCurIP];
      if (tt = TK_KEYWORD) and (fCurFunc[fCurIP][1] = '(') then
      begin
        // (expr)
        Inc(fCurIP); result := Expression(); if fError then exit;
        if (fCurFunc.Objects[fCurIP] <> TK_KEYWORD) or
           (fCurFunc[fCurIP][1] <> ')') then
        begin
          Error('")" expected');
          result[0] := #0;
        end
        else Inc(fCurIP);
      end
      else if tt = TK_OTHER then
      begin
        result := fCurFunc[fCurIP]; Inc(fCurIP);
      end
      else if tt = TK_ID then
      begin
        result := fCurFunc[fCurIP]; Inc(fCurIP);
        // do not dereference train leaders
        if (fCurFunc.Objects[fCurIP] <> TK_KEYWORD) or (fCurFunc[fCurIP][1] <> '.') then
        begin
          // dereference variable
          result := GetLocalVar(result);
        end;
      end
      else if (fCurFunc.Objects[fCurIP] = TK_KEYWORD) and
              (fCurFunc[fCurIP][1] in [KW_BITNOT,KW_LOGNOT]) then
      begin
        ch := fCurFunc[fCurIP][1]; Inc(fCurIP); // skip this token
        r0 := DoExpr(1); if fError then exit;
        result := fStdMath[ch](r0, '');
        //result := DoUFunc(ch, [result]);
      end
      else
      begin
        Error('invalid term');
        result[0] := #0;
      end;
      exit;
    end;

    // get first operand
    result := DoExpr(prio-1); if fError then exit;

    if prio = 1 then
    begin
      // priority level 1 -- opertions like "." and "[]", function calls
      while (fCurFunc.Objects[fCurIP] = TK_KEYWORD) and
            (Length(fCurFunc[fCurIP]) = 1) do
      begin
        ch := fCurFunc[fCurIP][1]; Inc(fCurIP); // skip this token
        case ch of
          '(': // function call
            begin
              (*if (result[0] <> #0) and not URALIsValidId(result) then
              begin
                Error('invalid function name: "'+result+'"');
                result[0] := #0;
                exit;
              end;*)
              funcname := result; URALUpStr(funcname); // save name
              result[0] := #0;
              // collect args
              al := TURALStrList.Create(false);
              try
                ParseFunctionArgs(al); if fError then exit;
                if funcname[0] = #0 then result := DoExecute(funcname, al)
                else if funcname = 'EVALEXPR' then result := DoEvalExpr(al)
                else if funcname = 'EVALSUB' then result := DoEvalSub(al)
                else
                begin
                  {$IFDEF XCOMPILE}
                  if (Length(funcname) > 4) and (funcname[1] = #0) then
                  begin
                    Move(funcname[2], f, 4);
                    uf := TURALFunction(fExtFuncs.Objects[f]);
                  end
                  else uf := FindExternalFunc(funcname);
                  {$ELSE}
                  uf := FindExternalFunc(funcname);
                  {$ENDIF}
                  if uf <> nil then result := uf.ExecuteEx(self, al, funcname)
                  else result := DoExecute(funcname, al);
                end;
                if fError then result[0] := #0;
              finally
                al.Free();
              end;
            end;
          '[': // string indexing
              DoStringIndexing(result);
          '.':
            begin
              r0 := result; result := CollectTrain(r0);
              if fError then exit;
              result := GetLocalVar(result);
            end;
          else Dec(fCurIP); break; // get back to the used token
        end;
        if fError then exit;
      end;
      // no other operators can be found at this priority level
      exit;
    end;

    // prority level 2 or more
    repeat
      f := 0;
      // is next token a possible math operation?
      if (fCurFunc.Objects[fCurIP] <> TK_KEYWORD) or
         (Length(fCurFunc[fCurIP]) <> 1) then break; // not a math token anyway
      ch := fCurFunc[fCurIP][1];
      while f <= MAX_OP_AT_LEVEL do
      begin
        opch := MATH_OPZ[prio, f]; Inc(f);
        if opch = #0 then begin f := MAX_OP_AT_LEVEL+1; break; end;
        if ch <> opch then continue;
        Inc(fCurIP); // skip operator token
        if prio = PRIO_RIGHT then
        begin
          // right associative
          r0 := DoExpr(prio);
        end
        else
        begin
          // normal
          r0 := DoExpr(prio-1);
        end;
        //result := DoUFunc(ch, [result, r0]);
        t := result; result := fStdMath[ch](t, r0);
        if fError then begin result[0] := #0; exit; end;
        break;
      end;
    until f > MAX_OP_AT_LEVEL; // another priority level?
  end;

var
  neg: Boolean;
  t: ShortString;
begin
  result[0] := #0; neg := false;
  if fCurFunc.Objects[fCurIP] = TK_KEYWORD then
  begin
    if fCurFunc[fCurIP][1] = KW_MINUS then
    begin
      neg := true;
      Inc(fCurIP);
    end
    else if fCurFunc[fCurIP][1] = KW_PLUS then Inc(fCurIP);
  end;
  result := DoExpr(MAX_PRIO); if fError then begin result[0] := #0; exit; end;
  if neg then
  begin
    t := result; result := fStdMath[KW_MINUS]('0', t);
  end;
end;

function TURALInterpreter.IsGlobalExists (const pName: ShortString): Boolean;
begin
  result := false;
end;

function TURALInterpreter.GetGlobalVar (const pName: ShortString): ShortString;
begin
  result := pName;
end;

procedure TURALInterpreter.SetGlobalVar (const pName, value: ShortString);
begin
end;

function TURALInterpreter.IsGlobalExistsCI (pName: ShortString): Boolean;
begin
  URALUpStr(pName);
  result := IsGlobalExists(pName);
end;

function TURALInterpreter.GetGlobalVarCI (pName: ShortString): ShortString;
begin
  URALUpStr(pName);
  result := GetGlobalVar(pName);
end;

procedure TURALInterpreter.SetGlobalVarCI (pName: ShortString; const value: ShortString);
begin
  URALUpStr(pName);
  SetGlobalVar(pName, value);
end;


function TURALInterpreter.GetLocalVar (const pName: ShortString): ShortString;
var
  f: Integer;
begin
  if pName = 'PI' then begin result := '3.141592653589793'; exit; end;
  if IsGlobalExists(pName) then begin result := GetGlobalVar(pName); exit; end;
  f := fVarPoolUsed-1;
  while (f >= fCurPoolTop) and (fVarPool[f].name <> pName) do Dec(f);
  if f >= fCurPoolTop then result := fVarPool[f].value else result := pName;
end;

procedure TURALInterpreter.SetLocalVar (const pName, value: ShortString);
var
  f: Integer;
begin
  if pName = 'PI' then exit;
  if IsGlobalExists(pName) then begin SetGlobalVar(pName, value); exit; end;
  f := fVarPoolUsed-1;
  while (f >= fCurPoolTop) and (fVarPool[f].name <> pName) do Dec(f);
  if f >= fCurPoolTop then
  begin
    fVarPool[f].value := value;
  end
  else
  begin
    if fVarPoolUsed = Length(fVarPool) then
    begin
      SetLength(fVarPool, fVarPoolUsed+64);
    end;
    fVarPool[fVarPoolUsed].name := pName;
    fVarPool[fVarPoolUsed].value := value;
    Inc(fVarPoolUsed);
  end;
end;

function TURALInterpreter.GetLocalVarCI (pName: ShortString): ShortString;
begin
  URALUpStr(pName);
  result := GetLocalVar(pName);
end;

procedure TURALInterpreter.SetLocalVarCI (pName: ShortString; const value: ShortString);
begin
  URALUpStr(pName);
  SetLocalVar(pName, value);
end;

function TURALInterpreter.Execute (pFuncName: ShortString; const args: array of ShortString): ShortString;
var
  al: TURALStrList;
  f: Integer;
begin
  al := TURALStrList.Create();
  try
    for f := 0 to High(args) do al.Add(args[f]);
    result := Execute(pFuncName, al);
  finally
    al.Free();
  end;
end;

function TURALInterpreter.Execute (pFuncName: ShortString; args: TURALStrList): ShortString;
begin
  if fInDoExecute then
  begin
    result := DoExecute(pFuncName, args);
    exit;
  end;

  result[0] := #0;
  fVarPoolUsed := 0; fVarPoolMax := 0; fCurPoolTop := 0;
  fInstructionCount := 0;
  fCurArgs := nil; fCurFuncName[0] := #0;
  fCurFunc := nil; fCurIP := 0;
  try
    fInDoExecute := true;
    result := DoExecute(pFuncName, args);
  finally
    fInDoExecute := false;
    // shrink variable pool
    if fVarPoolMax < 256 then fVarPoolMax := 256;
    if Length(fVarPool) > fVarPoolMax then SetLength(fVarPool, fVarPoolMax+64);
  end;
end;

function TURALInterpreter.EvalSub (pSubBody, pArgs: TURALStrList): ShortString;
var
  fn: TURALStrList;
begin
  PrecompileFunction(pSubBody, fn);
  if fn = nil then
  begin
    Error('EXECSUB() can''t prepare sub');
    result[0] := #0;
    exit;
  end;
  try
    if fInDoExecute then result := DoExecute('', pArgs, fn)
    else
    begin
      fError := false; fErrorMsg[0] := #0;
      fVarPoolUsed := 0; fVarPoolMax := 0; fCurPoolTop := 0;
      fInstructionCount := 0;
      fCurArgs := nil; fCurFuncName[0] := #0;
      fCurFunc := nil; fCurIP := 0;
      try
        fInDoExecute := true;
        result := DoExecute('', pArgs, fn);
      finally
        fInDoExecute := false;
        // shrink variable pool
        if fVarPoolMax < 256 then fVarPoolMax := 256;
        if Length(fVarPool) > fVarPoolMax then SetLength(fVarPool, fVarPoolMax+64);
      end;
    end;
  finally
    fn.Free();
  end;
end;

function TURALInterpreter.EvalSub (pSubBody: TURALStrList; pArgs: array of ShortString): ShortString;
var
  al: TURALStrList;
  f: Integer;
begin
  al := TURALStrList.Create(false);
  try
    for f := 0 to High(pArgs) do al.Add(pArgs[f]);
    result := EvalSub(pSubBody, al);
  finally
    al.Free();
  end;
end;

function TURALInterpreter.EvalSub (const pSubBody: ShortString; pArgs: TURALStrList): ShortString;
var
  fn: TURALStrList;
begin
  fn := ShortStringToStrList(pSubBody);
  try
    result := EvalSub(fn, pArgs);
  finally
    fn.Free();
  end;
end;

function TURALInterpreter.EvalSub (const pSubBody: ShortString; pArgs: array of ShortString): ShortString;
var
  al: TURALStrList;
  f: Integer;
begin
  al := TURALStrList.Create(false);
  try
    for f := 0 to High(pArgs) do al.Add(pArgs[f]);
    result := EvalSub(pSubBody, al);
  finally
    al.Free();
  end;
end;

function TURALInterpreter.EvalSubLong (const pSubBody: AnsiString; pArgs: TURALStrList): ShortString;
var
  fn: TURALStrList;
begin
  fn := AnsiStringToStrList(pSubBody);
  try
    result := EvalSub(fn, pArgs);
  finally
    fn.Free();
  end;
end;

function TURALInterpreter.EvalSubLong (const pSubBody: AnsiString; pArgs: array of ShortString): ShortString;
var
  al: TURALStrList;
  f: Integer;
begin
  al := TURALStrList.Create(false);
  try
    for f := 0 to High(pArgs) do al.Add(pArgs[f]);
    result := EvalSub(pSubBody, al);
  finally
    al.Free();
  end;
end;

function TURALInterpreter.DoExecute (pFuncName: ShortString; args: TURALStrList;
  thisSub: TURALStrList = nil): ShortString;
// if thisSub <> nil, do not search sub in sublist, just execute it
const
  MAX_WHILE_NESTING = 32;
label
  findelse, specialexpression;
var
  oa, ofn: TURALStrList;
  oip, optop, ip, iplast: Integer;
  oname: ShortString;
  fn: TURALStrList;
  f: Integer;
  pint: PInteger;
  whileStack: array [0..MAX_WHILE_NESTING-1] of Integer; // ip of "WHILE" condition
  ifCnt: Integer;
  whileCnt: Integer;
  t: ShortString;

  function FindNextElse (): Boolean;
  // false: no more ELSEs
  // end token NOT skipped
  begin
    result := false;
    f := 0;
    while fCurIP < iplast do
    begin
      if (fn.Objects[fCurIP] = TK_KEYWORD) and (Length(fn[fCurIP]) = 1) then
      begin
        case fn[fCurIP][1] of
          KW_IF: Inc(f);
          KW_ELSE, KW_ELSIF, KW_ENDIF:
            if f = 0 then
            begin
              result := fn[fCurIP][1] <> KW_ENDIF; break;
            end
            else Dec(f);
        end;
      end;
      Inc(fCurIP);
    end;
  end;

  function FindOperatorEnd (bt, et: Char): Boolean;
  // false: not found
  // end token skipped
  var
    ch: Char;
  begin
    result := false;
    f := 0;
    while fCurIP < iplast do
    begin
      if (fn.Objects[fCurIP] = TK_KEYWORD) and (Length(fn[fCurIP]) = 1) then
      begin
        ch := fn[fCurIP][1];
        if ch = bt then Inc(f)
        else if ch = et then
        begin
          if f = 0 then begin result := true; Inc(fCurIP); break; end else Dec(f);
        end;
      end;
      Inc(fCurIP);
    end;
  end;

begin
  fError := false; fErrorMsg[0] := #0;
  if not fInDoExecute then
  begin
    if thisSub <> nil then
    begin
      Error('can''t call nameless function outside of DoExecute() loop');
      exit;
    end;
    result := Execute(pFuncName, args);
    exit;
  end;

  result[0] := #0; if fError then exit;
  if thisSub = nil then
  begin
    if pFuncName[0] = #0 then
    begin
      if fCurFunc = nil then
      begin
        Error('can''t directly call nameless function');
        exit;
      end;
      fn := fCurFunc;
    end
    else
    begin
      URALUpStr(pFuncName);
      fn := TURALStrList(fFuncList.Objects[fFuncList.IndexOf(pFuncName)]);
      if fn = nil then
      begin
        Error('unknown function: "'+pFuncName+'"');
        exit;
      end;
    end;
  end
  else fn := thisSub;

  oa := fCurArgs; ofn := fCurFunc; oip := fCurIP; oname := fCurFuncName;
  optop := fCurPoolTop;
  try
    Inc(PInteger(@(fn.fItems^[0]))^); // increment execution counter

    if args = nil then
    begin
      fCurArgs := TURALStrList.Create(); ifCnt := 0;
    end
    else
    begin
      fCurArgs := args; ifCnt := args.Count;
    end;
    fCurFuncName := pFuncName;
    fCurFunc := fn; fCurPoolTop := fVarPoolUsed;
    SetLocalVar('RESULT', '');
    // now assign arguments
    // the header was prepared by AddFunction()
    ip := 1; f := 0;
    while ip < fn.Count do
    begin
      if fn.Objects[ip] = TK_EOP then break;
      if fn.Objects[ip] <> TK_ID then
      begin
        Error('internal bug: invalid header in "'+pFuncName+'"');
        result[0] := #0;
        exit;
      end;
      if f >= ifCnt then SetLocalVar(fn[ip], '')
      else begin SetLocalVar(fn[ip], args[f]); Inc(f); end;
      Inc(ip);
    end;
    Inc(ip); // skip EOL
    fCurIP := ip;
    whileCnt := 0; ifCnt := 0;
    Inc(fInstructionCount); iplast := fn.Count;
    while (fCurIP < iplast) and not fError do
    begin
      if fInstructionCount >= fInstructionLimit then
      begin
        Error('instruction limit exceeded');
        break;
      end;
      if fn.Objects[fCurIP] = TK_ID then
      begin
        // function call or assignment
        Inc(fInstructionCount);
        Inc(fCurIP);
        if fn.Objects[fCurIP] = TK_KEYWORD then
        begin
          // simple assign or function call
          // (simple instructions executes faster)
          f := fCurIP-1; // name position
          case fn[fCurIP][1] of
            KW_EQU:
              begin
                Inc(fCurIP); result := Expression(); if fError then break;
                {$IFDEF XCOMPILE}
                t := fn[f];
                if (Length(t) > 4) and (t[1] = #0) then Delete(t, 1, 5);
                SetLocalVar(t, result);
                {$ELSE}
                SetLocalVar(fn[f], result);
                {$ENDIF}
              end;
            '(', '.':
              begin
                t := fn[fCurIP-1];
                goto specialexpression;
              end;
            else Error('bad synthax'); break;
          end;
        end
        else
        begin
          t := fn[fCurIP-1];
specialexpression:
          repeat
            if fn.Objects[fCurIP] <> TK_KEYWORD then
            begin
              Dec(fCurIP);
              Error('bad synthax');
              break;
            end;
            case fn[fCurIP][1] of
              KW_EQU:
                begin
                  Inc(fCurIP); result := Expression(); if fError then break;
                  t := GetLocalVar(t); URALUpStr(t);
                  SetLocalVar(t, result);
                  break;
                end;
              '(': // function call
                begin
                  Inc(fCurIP); result := DoFunctionCall(t); if fError then break;
                  t := result;
                end;
              '.':
                begin
                  Inc(fCurIP); result := CollectTrain(t); if fError then break;
                  t := result;
                end;
              else break;
            end;
          until fn.Objects[fCurIP] = TK_EOP;
          //if fn.Objects[fCurIP] <> TK_EOP then begin Error('bad synthax'); break; end;
        end;
      end
      else if (fn.Objects[fCurIP] = TK_KEYWORD) and (Length(fn[fCurIP]) = 1) then
      begin
        Inc(fInstructionCount); Inc(fCurIP);
        case fn[fCurIP-1][1] of
          KW_SUBEND: break;
          KW_IF:
            begin
              Inc(ifCnt);
              // check condition
              result := Expression(); if fError then break;
              // skip possible THEN
              if fn.Objects[fCurIP] = TK_EOP then Inc(fCurIP)
              else if (fn.Objects[fCurIP] = TK_KEYWORD) and
                      (fn[fCurIP] = KW_THEN) then Inc(fCurIP)
              else
              begin
                Error('bad IF synthax');
                break;
              end;
              if not URALIsTrue(result) then
              begin
                // skip IF
                {Dec(ifCnt);
                if not FindOperatorEnd(KW_IF, KW_ENDIF) then
                begin
                  Error('IF without ENDIF');
                  break;
                end;}
                goto findelse;
              end;
            end;
          KW_THEN:
            begin
              Error('THEN without IF');
              break;
            end;
          KW_GOELSE:
            begin
              if ifCnt = 0 then
              begin
                Error('GOELSE without IF');
                break;
              end;
findelse:     repeat
                if not FindNextElse() then
                begin
                  if (fn.Objects[fCurIP] <> TK_KEYWORD) or
                     (fn[fCurIP] <> KW_ENDIF) then
                  begin
                    Error('IF without ENDIF');
                    break;
                  end;
                  Dec(ifCnt); Inc(fCurIP);
                  break;
                end
                else
                begin
                  if fn[fCurIP][1] = KW_ELSIF then
                  begin
                    Inc(fCurIP);
                    // check condition
                    result := Expression(); if fError then break;
                    // skip possible THEN
                    if fn.Objects[fCurIP] = TK_EOP then Inc(fCurIP)
                    else if (fn.Objects[fCurIP] = TK_KEYWORD) and
                            (fn[fCurIP] = KW_THEN) then Inc(fCurIP)
                    else
                    begin
                      Error('bad IF synthax');
                      break;
                    end;
                    if not URALIsTrue(result) then continue; // find next...
                  end
                  else Inc(fCurIP);
                  break;
                end;
              until false;
              if fError then break;
            end;
          KW_ELSIF, KW_ELSE:
            begin
              if ifCnt = 0 then
              begin
                Error('ELSE without IF');
                break;
              end;
              Dec(ifCnt);
              if not FindOperatorEnd(KW_IF, KW_ENDIF) then
              begin
                Error('IF without ENDIF');
                break;
              end;
            end;
          KW_ENDIF:
            begin
              if ifCnt = 0 then
              begin
                Error('ENDIF without IF');
                break;
              end;
              Dec(ifCnt);
            end;
          KW_BREAKIF:
            begin
              if ifCnt = 0 then
              begin
                Error('BREAKIF without IF');
                break;
              end;
              Dec(ifCnt);
              if not FindOperatorEnd(KW_IF, KW_ENDIF) then
              begin
                Error('IF without ENDIF');
                break;
              end;
            end;
          KW_WHILE, KW_WEND, KW_AGAIN:
            begin
              if fn[fCurIP-1][1] = KW_WHILE then
              begin
                // save IP
                if whileCnt = MAX_WHILE_NESTING then
                begin
                  Error('too many nested WHILEs');
                  break;
                end;
                whileStack[whileCnt] := fCurIP; Inc(whileCnt);
              end
              else
              begin
                // start the loop again
                if whileCnt = 0 then
                begin
                  Error('WEND without WHILE');
                  break;
                end;
                fCurIP := whileStack[whileCnt-1];
              end;
              // check condition
              result := Expression(); if fError then break;
              if fn.Objects[fCurIP] <> TK_EOP then
              begin
                Error('bad WHILE synthax');
                break;
              end;
              if not URALIsTrue(result) then
              begin
                Dec(whileCnt); // delete top WHILE
                if not FindOperatorEnd(KW_WHILE, KW_WEND) then
                begin
                  Error('WHILE without WEND');
                  break;
                end;
              end;
            end;
          KW_BREAK:
            begin
              if whileCnt = 0 then
              begin
                Error('BREAK without WHILE');
                break;
              end;
              Dec(whileCnt); // delete top WHILE
              if not FindOperatorEnd(KW_WHILE, KW_WEND) then
              begin
                Error('WHILE without WEND');
                break;
              end;
            end;
          KW_RETURN:
            begin
              if fn.Objects[fCurIP] = TK_EOP then break;
              result := Expression();
              if not fError then SetLocalVar('RESULT', result);
              break;
            end;
          '(':
            begin
              // expression that starts with "("
              // checked here for speed reasons
              t := Expression(); if fError then break;
              if (fn.Objects[fCurIP] <> TK_KEYWORD) or (fn[fCurIP][1] <> ')') then
              begin
                Error('")" expected');
                break;
              end;
              Inc(fCurIP);
              if fn.Objects[fCurIP] <> TK_EOP then
                goto specialexpression; // jump to aviod code duplication
            end
          else Error('bad synthax'); break;
        end;
      end
      else if fn.Objects[fCurIP] = TK_EOP then Inc(fCurIP)
      else
      begin
        Error('bad synthax');
        result[0] := #0;
        break;
      end;
    end;
    if not fError then result := GetLocalVar('RESULT') else result[0] := #0;
  finally
    if args = nil then URALFreeAndNil(fCurArgs);
    fVarPoolUsed := fCurPoolTop;
    fCurArgs := oa; fCurFuncName := oname; fCurIP := oip; fCurFunc := ofn;
    fCurPoolTop := optop;

    // check for cached update
    pInt := PInteger(@(fn.fItems^[0])); // execution counter
    Dec(pInt^); // decrement execution counter
    if (thisSub = nil) and (pInt^ = 0) then
    begin
      if fn.fItems^[0][4] <> #0 then
      begin
        // delete this sub
        fn.Objects[0].Free(); fn.Objects[0] := nil;
        f := fFuncList.Count-1;
        while (f >= 0) and (fFuncList.Objects[f] <> fn) do Dec(f);
        if f <> -1 then fFuncList.Delete(f);
      end
      else if fn.Objects[0] <> nil then
      begin
        // update this sub
        thisSub := TURALStrList(fn.Objects[0]);
        f := fFuncList.Count-1;
        while (f >= 0) and (fFuncList.Objects[f] <> fn) do Dec(f);
        if f <> -1 then
        begin
          fFuncList.Objects[f].Free(); fFuncList.Objects[f] := thisSub;
        end;
      end;
    end;
  end;
end;


function TURALInterpreter.FunctionListSL (const txt: TURALStrList): TURALStrList;
var
  fn1: TURALStrList;
  f, e: Integer;
begin
  result := nil;
  fn1 := Tokenize(txt);
  if fn1 = nil then exit;
  result := TURALStrList.Create(false);
  try
    f := 0; while f < fn1.Count do
    begin
      while f < fn1.Count do
      begin
        if (fn1.Objects[f] = TK_KEYWORD) and (fn1[f][1] = KW_SUB) then break;
        Inc(f);
      end;
      if f > fn1.Count then break;
      e := f+1;
      while e < fn1.Count do
      begin
        if (fn1.Objects[e] = TK_KEYWORD) and (fn1[e][1] = KW_SUBEND) then begin Inc(e); break; end;
        Inc(e);
      end;
      f := e;
      result.Add(fn1[f+1]);
    end;
  except
    fn1.Free();
    result.Free();
    raise;
  end;
  fn1.Free();
end;

function TURALInterpreter.FunctionList (const txt: AnsiString): TURALStrList;
var
  sl: TURALStrList;
begin
  sl := SplitStringToList(txt);
  try
    result := FunctionListSL(sl);
  finally
    sl.Free();
  end;
end;

function TURALInterpreter.AddTextSL (const txt: TURALStrList): ShortString;
var
  fn, fn1: TURALStrList;
  f, e: Integer;
begin
  result := '';
  fn1 := Tokenize(txt);
  if fn1 = nil then exit;
  try
    f := 0; while f < fn1.Count do
    begin
      while f < fn1.Count do
      begin
        if (fn1.Objects[f] = TK_KEYWORD) and (fn1[f][1] = KW_SUB) then break;
        Inc(f);
      end;
      if f > fn1.Count then break;
      e := f+1;
      while e < fn1.Count do
      begin
        if (fn1.Objects[e] = TK_KEYWORD) and (fn1[e][1] = KW_SUBEND) then begin Inc(e); break; end;
        Inc(e);
      end;
      fn := TURALStrList.Create();
      try
        while f < e do
        begin
          fn.AddObject(fn1[f], fn1.Objects[f]); Inc(f);
        end;
        AddFunction(fn, true);
        if ErrorFlag then
        begin
          result := fn1[f+1]+' ('+URALInt2Str(ErrorLine)+'): '+ErrorMsg;
          exit;
        end;
      finally
        fn.Free();
      end;
    end;
  finally
    fn1.Free();
  end;
end;

function TURALInterpreter.AddText (const txt: AnsiString): ShortString;
var
  sl: TURALStrList;
begin
  sl := SplitStringToList(txt);
  try
    result := AddTextSL(sl);
  finally
    sl.Free();
  end;
end;


function SplitStringToList (const txt: AnsiString): TURALStrList;
var
  f, e: Integer;
  s: ShortString;
begin
  result := TURALStrList.Create(false);
  try
    f := 1;
    while (f <= Length(txt)) do
    begin
      while (f <= Length(txt)) and (txt[f] in [#0..#9,#11,#12,#14..' ']) do Inc(f);
      if f > Length(txt) then break;
      e := f; while (e <= Length(txt)) and not (txt[e] in [#10, #13]) do Inc(e);
      s := Copy(txt, f, e-f);
      result.Add(s);
      f := e; if f > Length(txt) then break;
      Inc(f);
      if txt[f-1] = #10 then
      begin
        if f > Length(txt) then break;
        if txt[f] = #13 then Inc(f);
      end;
    end;
  except
    result.Free();
    raise;
  end;
end;


{$I-}
function LoadTextFileToList (const fName: ShortString): TURALStrList;
var
  fi: Text;
  s: ShortString;
begin
  result := nil;
  IOResult();
  AssignFile(fi, fName);
  Reset(fi); if IOResult() <> 0 then exit;
  try
    result := TURALStrList.Create(false);
    while not EOF(fi) do
    begin
      ReadLn(fi, s);
      result.Add(s);
    end;
  except
    CloseFile(fi);
    result.Free();
    raise;
  end;
  CloseFile(fi);
end;


end.
