// original code: WDOSX-Pack v1.07, (c) 1999-2001 by Joergen Ibsen / Jibz
// for data and executable compression software:
//   http://apack.cjb.net/   mailto:jibz@hotmail.com
//
// Delphi translation: Ketmar // Vampire Avalon
//
// procedures should be safe...
//
unit wpackU;

interface


function WDXPack (const source; length: Integer;
  out destination; destSize: Integer): Integer;
  // uncompressible data may be expanded by up to 1 bit per byte.
  // this means that the destination should be length+((length+7)/8)+2 bytes.
  // returns size of the packed data or -1 on error
function WDXUnpack (const source; srcLength: Integer;
  out destination; length: Integer): Integer;
  // unpack length bytes from source to destination
  // returns # of used bytes in source
  // destination buffer must be at least length bytes
  // returns -1 on error (out of source buffer)


implementation


const
  LEN_2_LIMIT = 1920;


type
  PInt = ^Integer;
  PIntArr = ^TIntArr;
  TIntArr = packed array [0..(maxInt div 4)-128] of Integer;

  TMatch = record
    pos: Integer;
    len: Integer;
  end;


function WDXPack (const source; length: Integer;
  out destination; destSize: Integer): Integer;
  // global variables
  var
    nextBackEntry: Integer;
    backPtr, sourceOfs: PAnsiChar;
    inBuffer, outBuffer, tagByte: PAnsiChar;
    bitCount: Integer;
    outEnd: Integer; // end if output buffer

    // back-table
    backTable: PIntArr;

    // lookup-table
    lookup: packed array [0..255, 0..255] of Integer;

    lastMatchPos: Integer;
    lastWasMatch: Boolean;
    error: Boolean;


  procedure AdvanceTagByte (bit: Boolean);
  var
    b: Byte;
  begin
    if error then exit;
    // check bitcount and then decrement
    if bitCount = 0 then
    begin
      bitCount := 8;
      tagByte := outBuffer; Inc(outBuffer); //tagByte^ := #0;
    end;
    if Integer(tagByte) >= outEnd then begin error := true; exit; end;
    // shift in bit
    b := Byte(tagByte^) shl 1; if bit then b := b or $01;
    tagByte^ := Char(b);
    Dec(bitCount);
  end;

  // output Gamma2-code for val in range [2..?] ...
  procedure OutputGamma (val: Integer);
  var
    invertlen, invert: Integer;
  begin
    invertlen := 0; invert := 0;
    // rotate bits into invert (except last)
    repeat
      invert := (invert shl 1) or (val and $0001);
      Inc(invertlen);
      val := (val shr 1) and $7FFFFFFF;
    until not (val > 1);
    //while ((val >>= 1) > 1);

    // output Gamma2-encoded bits
    Dec(invertlen);
    while invertlen > 0 do
    begin
      AdvanceTagByte((invert and $01) <> 0);
      AdvanceTagByte(true);
      invert := invert shr 1;
      Dec(invertlen);
    end;
    AdvanceTagByte((invert and $01) <> 0);
    AdvanceTagByte(false);
  end;

  procedure OutputLiteral (lit: Char);
  begin
    //WriteLn('lit');
    lastWasMatch := false;
    AdvanceTagByte(false); // FALSE indicates a literal
    if error then exit;
    if Integer(outBuffer) >= outEnd then begin error := true; exit; end;
    outBuffer^ := lit; Inc(outBuffer); // output the literal
  end;

  procedure OutputCodePair (pos, len: Integer; buffer: PAnsiChar);
  begin
    //WriteLn(pos, ':', len);
    // if we just had a match, don't use lastMatchPos
    if lastWasMatch then
    begin
      // if a short match is too far away, encode it as two literals instead
      if (pos > LEN_2_LIMIT) and (len = 2) then
      begin
        OutputLiteral(buffer^);
        OutputLiteral((buffer+1)^);
      end
      else
      begin
        AdvanceTagByte(true); // TRUE indicates a match
        // a match more than LEN_2_LIMIT bytes back will be longer than 2
        if pos > LEN_2_LIMIT then Dec(len);
        OutputGamma(len); // output length
        lastMatchPos := pos; Dec(pos);
        ASSERT(pos >= 0);
        OutputGamma(((pos shr 6) and $3FFFFFFF)+2); // output high part of position
        // output low 6 bits of position
        AdvanceTagByte((pos and $20) <> 0);
        AdvanceTagByte((pos and $10) <> 0);
        AdvanceTagByte((pos and $08) <> 0);
        AdvanceTagByte((pos and $04) <> 0);
        AdvanceTagByte((pos and $02) <> 0);
        AdvanceTagByte((pos and $01) <> 0);
      end
    end
    else
    begin
      lastWasMatch := true;
      // if a short match is too far away, encode it as two literals instead
      if (pos > LEN_2_LIMIT) and (len = 2) and (pos <> lastMatchPos) then
      begin
        OutputLiteral(buffer^);
        OutputLiteral((buffer+1)^);
      end
      else
      begin
        // TRUE indicates a match
        AdvanceTagByte(true);
        // a match more than LEN_2_LIMIT bytes back will be longer than 2
        if (pos > LEN_2_LIMIT) and (pos <> lastMatchPos) then Dec(len);
        OutputGamma(len); // output length
        // output position
        if pos = lastMatchPos then
        begin
          // a match with position 0 means use last position
          AdvanceTagByte(false);
          AdvanceTagByte(false);
        end
        else
        begin
          lastMatchPos := pos; Dec(pos);
          ASSERT(pos >= 0);
          OutputGamma(((pos shr 6) and $3FFFFFFF)+3); // output high part of position
          // output low 6 bits of position
          AdvanceTagByte((pos and $20) <> 0);
          AdvanceTagByte((pos and $10) <> 0);
          AdvanceTagByte((pos and $08) <> 0);
          AdvanceTagByte((pos and $04) <> 0);
          AdvanceTagByte((pos and $02) <> 0);
          AdvanceTagByte((pos and $01) <> 0);
        end;
      end;
    end;
  end;

  procedure OutputEOD ();
  begin
    // just switch last tagByte into position
    if error then exit;
    if bitCount = 8 then exit;
    tagByte^ := Char(Byte(tagByte^) shl bitCount);
  end;

  procedure FindMatch (var thematch: TMatch; buffer: PAnsiChar; lookback, lookforward: Integer);
  var
    backPos, matchLen, bestMatchLen, bestMatchPos: Integer;
    ptr: PAnsiChar;
    i0, i1: Integer;
    idx0, idx1: Integer;
  begin
    // temporary variables to avoid indirect addressing into the match
    bestMatchLen := 0; bestMatchPos := 0;

    // update lookup- and backtable up to current position
    while backPtr < buffer do
    begin
      idx0 := Byte(backPtr^);
      idx1 := Byte(PAnsiChar(backPtr+1)^);
      //PInt(Integer(backTable)+nextBackEntry*4)^ := lookup[idx0, idx1];
      backTable^[nextBackEntry] := lookup[idx0, idx1];
      lookup[idx0, idx1] := nextBackEntry;
      Inc(nextBackEntry); Inc(backPtr);
    end;

    // get position by looking up next two bytes
    backPos := lookup[Byte(buffer^), Byte(PAnsiChar(buffer+1)^)];

    if (backPos <> 0) and (lookforward > 1) then
    begin
      ptr := backPos+sourceOfs;

      // go backwards until before buffer
      while (ptr >= buffer) and (backPos <> 0) do
      begin
        //backPos := PInt(Integer(backTable)+backPos*4)^;
        backPos := backTable^[backPos];
        ptr := backPos+sourceOfs;
      end;

      // search through table entries
      while (backPos <> 0) and (buffer-ptr <= lookback) do
      begin
        matchLen := 2;
        // if this position has a chance to be better
        if PAnsiChar(ptr+bestMatchLen)^ = PAnsiChar(buffer+bestMatchLen)^ then
        begin
          // scan it
          while (PAnsiChar(ptr+matchLen)^ = PAnsiChar(buffer+matchLen)^) and
                (matchLen < lookforward) do Inc(matchLen);
          // check it
          if buffer-ptr = lastMatchPos then i0 := 1 else i0 := 0;
          if bestMatchPos = lastMatchPos then i1 := 1 else i1 := 0;
          if matchLen+i0 > bestMatchLen+i1 then
          begin
            bestMatchLen := matchLen;
            if bestMatchLen = lookforward then backPos := 0;
            bestMatchPos := buffer-ptr;
          end;
        end;
        // move backwards to next position
        backPos := PInt(Integer(backTable)+backPos*4)^;
        ptr := backPos+sourceOfs;
      end; // while ((backPos) && (buffer - ptr <= lookback))
    end; // if ((backPos) && (lookforward > 1))

    // forget match if too far away
    if (bestMatchPos > LEN_2_LIMIT) and
       (bestMatchLen = 2) and
       (bestMatchPos <> lastMatchPos) then
    begin
      bestMatchLen := 0; bestMatchPos := 0;
    end;

    // update the match with best match
    thematch.len := bestMatchLen;
    thematch.pos := bestMatchPos;
  end;

var
  match, nextmatch, literalmatch, testmatch: TMatch;
  pos, lastpos, literalCount: Integer;
  i, j: Integer;
  i0, i1: Integer;
  workMem: Pointer;
begin
  if length < 1 then begin result := 0; exit; end;
  if destSize < 1 then begin result := -1; exit; end;
  GetMem(workMem, (length+4)*4);
  error := false;

  try
    inBuffer := @source; outBuffer := @destination;
    outEnd := Integer(outBuffer)+destSize;
    sourceOfs := inBuffer-1;
    backTable := Pointer(workMem);

    // init lookup- and backtable
    for i := 0 to 255 do for j := 0 to 255 do lookup[i][j] := 0;

    backPtr := inBuffer; backTable^[0] := 0; nextBackEntry := 1;

    lastpos := -1; lastMatchPos := -1;
    lastWasMatch := false; literalCount := 0;

    // the first byte is sent verbatim
    outBuffer^ := inBuffer^;
    Inc(outBuffer); Inc(inBuffer);

    // init tag-byte
    bitCount := 8; tagByte := outBuffer; Inc(outBuffer);

    // pack data
    pos := 1;
    while pos < length do
    begin
      // find best match at current position (if not already found)
      if pos = lastpos then
      begin
        match.len := nextmatch.len; match.pos := nextmatch.pos;
      end
      else FindMatch(match, inBuffer, pos, length-pos);

      // if we found a match, find the best match at the next position
      if match.len <> 0 then
      begin
        FindMatch(nextmatch, inBuffer+1, pos+1, length-(pos+1));
        lastpos := pos+1;
      end
      else nextmatch.len := 0;

      // decide if we should output a match or a literal
      if match.pos = lastMatchPos then i0 := 1 else i0 := 0;
      if nextmatch.pos = lastMatchPos then i1 := 1 else i1 := 0;
      if (match.len <> 0) and (match.len+i0 >= nextmatch.len+i1) then
      begin
        // output any pending literals
        if literalCount <> 0 then
        begin
          if literalCount = 1 then
          begin
            OutputLiteral(PAnsiChar(inBuffer-1)^);
          end
          else
          begin
            // check if there is a closer match with the required length
            FindMatch(testmatch, inBuffer-literalCount, literalmatch.pos, literalCount);
            if testmatch.len >= literalCount then
            begin
              OutputCodePair(testmatch.pos, literalCount, inBuffer-literalCount);
            end
            else
            begin
              OutputCodePair(literalmatch.pos, literalCount, inBuffer-literalCount);
            end;
          end;
          literalCount := 0;
        end;
        // output match
        OutputCodePair(match.pos, match.len, inBuffer);
        Inc(inBuffer, match.len);
        Inc(pos, match.len);
      end
      else
      begin
        // if ((match.len) && ( ... )
        // check if we are allready collecting literals
        if literalCount <> 0 then
        begin
          // if so, continue..
          Inc(literalCount);
          // have we collected as many as possible?
          if literalCount = literalmatch.len then
          begin
            OutputCodePair(literalmatch.pos, literalCount, inBuffer-literalCount+1);
            literalCount := 0;
          end;
        end
        else
        begin
          // if (literalCount)
          // if we had a match which was not good enough, then save it..
          if match.len <> 0 then
          begin
            literalmatch.len := match.len;
            literalmatch.pos := match.pos;
            Inc(literalCount);
          end
          else
          begin
            // if not, we have to output the literal now
            OutputLiteral(inBuffer^);
          end;
        end; // if (literalCount)
        Inc(inBuffer); Inc(pos);
      end; // if ((match.len) && ( ... )
    end;

    // output any remaining literal bytes
    if literalCount <> 0 then
    begin
      if literalCount = 1 then OutputLiteral(PAnsiChar(inBuffer-1)^)
      else OutputCodePair(literalmatch.pos, literalCount, inBuffer-literalCount);
      //literalCount := 0;
    end;

    // do EOD stuff
    OutputEOD();

    if error then result := -1
    else result := Integer(outBuffer)-Integer(@destination);
  finally
    FreeMem(workMem);
  end;
end;


function WDXUnpack (const source; srcLength: Integer;
  out destination; length: Integer): Integer;
var
  fbyte: Byte;
  src, dest, pp: PAnsiChar;
  lastMatchPos: Integer;
  lastWasMatch: Boolean;
  len, pos: Integer;
  error: Boolean;

  function GetBit (): Boolean;
  begin
    result := (fbyte >= $80); fbyte := (fbyte and $7F) shl 1;
    if error then exit;
    if fbyte = 0 then
    begin
      if srcLength < 1 then begin error := true; exit; end;
      fbyte := Byte(src^); Inc(src); Dec(srcLength);
      result := (fbyte >= $80); fbyte := ((fbyte and $7F) shl 1) or $01;
    end;
  end;

  function GetGamma (): Integer;
  begin
    result := 1;
    if error then exit;
    repeat
      Inc(result, result); //result := result shl 1;
      if GetBit() then Inc(result); //result := result or $01;
    until not GetBit();
  end;

  function GetLoPos (pos: Integer): Integer;
  // get 6 low bits of position
  var
    f: Integer;
  begin
    for f := 0 to 5 do
    begin
      Inc(pos, pos); //pos := pos shl 1;
      if GetBit() then Inc(pos); //pos := pos or $01;
    end;
    result := pos;
  end;

begin
  result := 0; if length < 1 then exit;
  if srcLength < 1 then begin result := -1; exit; end;
  src := @source; dest := @destination;
  // the first byte was sent verbatim

  dest^ := src^; Inc(dest); Inc(src); Dec(length); Dec(srcLength);
  if length < 1 then begin result := 1; exit; end;
  lastMatchPos := 0; lastWasMatch := false; fbyte := 0;

  error := false;
  repeat
    if not GetBit() then
    begin
      // literal
      //WriteLn('lit');
      if srcLength < 1 then begin error := true; break; end;
      dest^ := src^; Inc(dest); Inc(src); Dec(length); Dec(srcLength);
      lastWasMatch := false;
    end
    else
    begin
      // match
      len := GetGamma(); // get length
      if error then break;
      // same position as last match?
      if lastWasMatch then
      begin
        pos := GetGamma()-2;
        if error then break;
        pos := GetLoPos(pos)+1;
        if error then break;
        lastMatchPos := pos;
        if pos > LEN_2_LIMIT then Inc(len);
      end
      else
      begin
        lastWasMatch := true;
        pos := GetGamma()-2;
        if error then break;
        if pos = 0 then pos := lastMatchPos
        else
        begin
          pos := GetLoPos(pos-1)+1;
          if error then break;
          lastMatchPos := pos;
          if pos > LEN_2_LIMIT then Inc(len);
        end;
      end;
      //WriteLn(pos, ':', len);
      if error then break;
      // copy match
      pp := dest; Dec(pp, pos);
      while (len > 0) and (length > 0) do
      begin
        dest^ := pp^; Inc(dest); Inc(pp); Dec(length); Dec(len);
      end;
    end;
  until (length < 1) or error;

  if error then result := -1 else result := Integer(src)-Integer(@source);
end;


end.
